#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import pprint
from dataclasses import dataclass


@dataclass
class Status:
    def as_obj(self) -> dict:
        return self.__dict__

    def as_string(self) -> str:
        return pprint.pformat(self.as_obj(), width=150)

    def as_json(self) -> str:
        return json.dumps(self.as_obj())



# generated by datamodel-codegen:
#   filename:  schema/type/entityUsage.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import List

from pydantic import BaseModel, Extra, Field

from . import entityReference, usageDetails


class UsageDetailsOfAnEntity(BaseModel):
    class Config:
        extra = Extra.forbid

    entity: entityReference.EntityReference = Field(
        ..., description='Entity for which usage is returned.'
    )
    usage: List[usageDetails.TypeUsedToReturnUsageDetailsOfAnEntity] = Field(
        ..., description='List usage details per day.'
    )



# generated by datamodel-codegen:
#   filename:  schema/type/paging.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Extra, Field


class Paging(BaseModel):
    class Config:
        extra = Extra.forbid

    before: Optional[str] = Field(
        None,
        description='Before cursor used for getting the previous page (see API pagination for details).',
    )
    after: Optional[str] = Field(
        None,
        description='After cursor used for getting the next page (see API pagination for details).',
    )
    total: int = Field(
        ..., description='Total number of entries available to page through.'
    )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import logging
import pathlib

from metadata.config.common import ConfigModel
from metadata.ingestion.api.common import Entity, WorkflowContext
from metadata.ingestion.api.stage import Stage, StageStatus
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig

logger = logging.getLogger(__name__)


class FileStageConfig(ConfigModel):
    filename: str


class FileStage(Stage[Entity]):
    config: FileStageConfig
    status: StageStatus

    def __init__(
        self,
        ctx: WorkflowContext,
        config: FileStageConfig,
        metadata_config: MetadataServerConfig,
    ):
        super().__init__(ctx)
        self.config = config
        self.status = StageStatus()

        fpath = pathlib.Path(self.config.filename)
        self.file = fpath.open("w")
        self.wrote_something = False

    @classmethod
    def create(
        cls, config_dict: dict, metadata_config_dict: dict, ctx: WorkflowContext
    ):
        config = FileStageConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(ctx, config, metadata_config)

    def stage_record(self, record: Entity) -> None:
        json_record = json.loads(record.json())
        self.file.write(json.dumps(json_record))
        self.file.write("\n")
        self.status.records_status(record)

    def get_status(self):
        return self.status

    def close(self):
        self.file.close()



# generated by datamodel-codegen:
#   filename:  schema/entity/policies/lifecycle/deleteAction.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import Any, Optional, Union

from pydantic import BaseModel, Extra, Field, conint


class LifecycleDeleteAction1(BaseModel):
    class Config:
        extra = Extra.forbid

    daysAfterCreation: Optional[conint(ge=1)] = Field(
        None,
        description='Number of days after creation of the entity that the deletion should be triggered.',
    )
    daysAfterModification: Optional[conint(ge=1)] = Field(
        None,
        description='Number of days after last modification of the entity that the deletion should be triggered.',
    )


class LifecycleDeleteAction(BaseModel):
    class Config:
        extra = Extra.forbid

    __root__: Union[LifecycleDeleteAction1, Any, Any] = Field(
        ...,
        description='An action to delete or expire the entity.',
        title='LifecycleDeleteAction',
    )



# generated by datamodel-codegen:
#   filename:  schema/api/services/createDatabaseService.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Field, constr

from ...entity.services import databaseService
from ...type import jdbcConnection, schedule


class CreateDatabaseServiceEntityRequest(BaseModel):
    name: constr(min_length=1, max_length=128) = Field(
        ..., description='Name that identifies the this entity instance uniquely'
    )
    description: Optional[str] = Field(
        None, description='Description of Database entity.'
    )
    serviceType: databaseService.DatabaseServiceType
    jdbc: jdbcConnection.JdbcInfo
    ingestionSchedule: Optional[schedule.Schedule] = Field(
        None, description='Schedule for running metadata ingestion jobs'
    )



# generated by datamodel-codegen:
#   filename:  data/tags/userTags.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import Any

from pydantic import BaseModel, Field


class Model(BaseModel):
    __root__: Any = Field(
        ...,
        description='Tags related to describing User related data. These tags are used to label data assets to describe the user data in those assets. Example - a table column can be labeled with `User.PhoneNumber` tag. The associated `PII` and `PersonalData` tags are automatically applied. This is done to help users producing the data  focus on describing the data without being policy experts. The associated tags take care of applying classification tags automatically.',
    )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Optional

from pydantic import BaseModel

from metadata.generated.schema.entity.data.database import Database
from metadata.generated.schema.entity.data.location import Location
from metadata.generated.schema.entity.data.table import Table


class OMetaDatabaseAndTable(BaseModel):
    database: Database
    table: Table
    location: Optional[Location]



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict, List, Optional

from metadata.generated.schema.entity.data.table import ColumnJoins
from metadata.ingestion.models.json_serializable import JsonSerializable
from pydantic import BaseModel


class TableQuery(JsonSerializable):
    """ """

    def __init__(
        self,
        query: str,
        user_name: str,
        starttime: str,
        endtime: str,
        analysis_date: str,
        database: str,
        aborted: bool,
        sql: str,
    ) -> None:
        """ """
        self.query = query
        self.user_name = user_name
        self.starttime = starttime
        self.endtime = endtime
        self.analysis_date = analysis_date
        self.database = database
        self.aborted = aborted
        self.sql = sql


class TableColumn(BaseModel):
    table: str
    column: str


class TableColumnJoin(BaseModel):
    table_column: Optional[TableColumn] = None
    joined_with: Optional[List[TableColumn]] = None


TableColumnJoins = List[TableColumnJoin]


class TableUsageCount(BaseModel):
    table: str
    date: str
    database: str
    count: int = 1
    joins: TableColumnJoins


class QueryParserData(BaseModel):
    tables: List[str]
    tables_aliases: Dict[str, str]
    columns: Dict[str, List[object]]
    date: str
    database: str
    sql: str

    class Config:
        arbitrary_types_allowed = True


class TableUsageRequest(BaseModel):
    date: str
    count: int


class ColumnJoinsList(BaseModel):
    __root__: List[ColumnJoins]


class ColumnJoinedWith(BaseModel):
    fullyQualifiedName: str
    joinCount: int


TablesUsage = List[TableUsageCount]



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import os
import uuid
from typing import Iterable, List

import boto3

from metadata.generated.schema.api.services.createStorageService import (
    CreateStorageServiceEntityRequest,
)
from metadata.generated.schema.type.entityReference import EntityReference
from metadata.ingestion.api.common import ConfigModel, Entity, WorkflowContext
from metadata.ingestion.api.source import Source, SourceStatus
from metadata.ingestion.ometa.ometa_api import OpenMetadata
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.generated.schema.entity.data.location import Location, LocationType
from metadata.generated.schema.entity.policies.filters import Filters1, Prefix
from metadata.generated.schema.entity.policies.policy import Policy, PolicyType
from metadata.generated.schema.entity.policies.lifecycle.rule import LifecycleRule
from metadata.generated.schema.entity.policies.lifecycle.moveAction import (
    Destination,
    LifecycleMoveAction,
)
from metadata.generated.schema.entity.services.storageService import StorageService
from metadata.generated.schema.type.storage import StorageServiceType, S3StorageClass

logger: logging.Logger = logging.getLogger(__name__)


class S3SourceConfig(ConfigModel):
    service_name: str
    aws_access_key_id: str
    aws_secret_access_key: str


class S3Source(Source[Entity]):
    config: S3SourceConfig
    status: SourceStatus

    def __init__(
        self, config: S3SourceConfig, metadata_config: MetadataServerConfig, ctx
    ):
        super().__init__(ctx)
        self.config = config
        self.metadata_config = metadata_config
        os.environ["AWS_ACCESS_KEY_ID"] = self.config.aws_access_key_id
        os.environ["AWS_SECRET_ACCESS_KEY"] = self.config.aws_secret_access_key
        self.status = SourceStatus()
        self.service = get_storage_service_or_create(
            config.service_name, metadata_config
        )
        self.s3 = boto3.resource("s3")

    @classmethod
    def create(
        cls, config_dict: dict, metadata_config_dict: dict, ctx: WorkflowContext
    ):
        config = S3SourceConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(config, metadata_config, ctx)

    def prepare(self):
        pass

    def next_record(self) -> Iterable[Entity]:
        try:
            for bucket in self.s3.buckets.all():
                self.status.scanned(bucket)
                bucket_name = self._get_bucket_name_with_prefix(bucket.name)
                location = Location(
                    id=uuid.uuid4(),
                    name=bucket_name,
                    displayName=bucket_name,
                    locationType=LocationType.Bucket,
                    service=EntityReference(
                        id=self.service.id,
                        type="storageService",
                        name=self.service.name,
                    ),
                )
                yield location

                # Retrieve lifecycle policies for the bucket.
                rules: List[LifecycleRule] = []
                for rule in self.s3.BucketLifecycleConfiguration(bucket.name).rules:
                    rules.append(self._get_rule(rule, location))
                policy_name = f"{bucket.name}-lifecycle-policy"
                yield Policy(
                    id=uuid.uuid4(),
                    name=policy_name,
                    displayName=policy_name,
                    description=policy_name,
                    policyType=PolicyType.Lifecycle,
                    rules=rules,
                    enabled=True,
                )
        except Exception as e:
            self.status.failure("error", str(e))

    def get_status(self) -> SourceStatus:
        return self.status

    @staticmethod
    def _get_bucket_name_with_prefix(bucket_name: str) -> str:
        return (
            "s3://" + bucket_name
            if not bucket_name.startswith("s3://")
            else bucket_name
        )

    def close(self):
        pass

    def _get_rule(self, rule: dict, location: Location) -> LifecycleRule:
        actions = []
        if "Transitions" in rule:
            for transition in rule["Transitions"]:
                if "StorageClass" in transition and "Days" in transition:
                    actions.append(
                        LifecycleMoveAction(
                            daysAfterCreation=transition["Days"],
                            destination=Destination(
                                storageServiceType=self.service,
                                storageClassType=S3StorageClass(
                                    transition["StorageClass"]
                                ),
                                location=location,
                            ),
                        )
                    )

        enabled = rule["Status"] == "Enabled" if "Status" in rule else False

        filters = []
        if "Filter" in rule and "Prefix" in rule["Filter"]:
            filters.append(Prefix.parse_obj(rule["Filter"]["Prefix"]))

        name = rule["ID"] if "ID" in rule else None

        return LifecycleRule(
            actions=actions,
            enabled=enabled,
            filters=Filters1.parse_obj(filters),
            name=name,
        )


def get_storage_service_or_create(
    service_name: str, metadata_config: MetadataServerConfig
) -> StorageService:
    metadata = OpenMetadata(metadata_config)
    service = metadata.get_by_name(entity=StorageService, fqdn=service_name)
    if service is not None:
        return service
    return metadata.create_or_update(
        CreateStorageServiceEntityRequest(
            name=service_name,
            serviceType=StorageServiceType.S3,
        )
    )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import csv
import json
import logging
import os
import uuid
from collections import namedtuple
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, List, Union

from pydantic import ValidationError

from metadata.config.common import ConfigModel
from metadata.generated.schema.api.data.createMlModel import CreateMlModelEntityRequest
from metadata.generated.schema.api.data.createTopic import CreateTopicEntityRequest
from metadata.generated.schema.api.lineage.addLineage import AddLineage
from metadata.generated.schema.entity.data.database import Database
from metadata.generated.schema.entity.data.location import Location, LocationType
from metadata.generated.schema.entity.data.mlmodel import MlModel
from metadata.generated.schema.entity.data.pipeline import Pipeline
from metadata.generated.schema.entity.data.table import Table
from metadata.generated.schema.entity.teams.user import User
from metadata.generated.schema.type.basic import Href
from metadata.generated.schema.type.entityLineage import EntitiesEdge
from metadata.generated.schema.type.entityReference import EntityReference
from metadata.ingestion.api.common import Entity
from metadata.ingestion.api.source import Source, SourceStatus
from metadata.ingestion.models.ometa_table_db import OMetaDatabaseAndTable
from metadata.ingestion.models.table_metadata import Chart, Dashboard
from metadata.ingestion.ometa.ometa_api import OpenMetadata
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.utils.helpers import (
    get_dashboard_service_or_create,
    get_database_service_or_create,
    get_database_service_or_create_v2,
    get_messaging_service_or_create,
    get_pipeline_service_or_create,
    get_storage_service_or_create,
)

logger: logging.Logger = logging.getLogger(__name__)

COLUMN_NAME = "Column"
KEY_TYPE = "Key type"
DATA_TYPE = "Data type"
COL_DESCRIPTION = "Description"
TableKey = namedtuple("TableKey", ["schema", "table_name"])


class InvalidSampleDataException(Exception):
    """
    Sample data is not valid to be ingested
    """


def get_lineage_entity_ref(edge, metadata_config) -> EntityReference:
    metadata = OpenMetadata(metadata_config)
    fqn = edge["fqn"]
    if edge["type"] == "table":
        table = metadata.get_by_name(entity=Table, fqdn=fqn)
        return EntityReference(id=table.id, type="table")
    elif edge["type"] == "pipeline":
        pipeline = metadata.get_by_name(entity=Pipeline, fqdn=fqn)
        return EntityReference(id=pipeline.id, type="pipeline")
    elif edge["type"] == "dashboard":
        dashboard = metadata.get_by_name(entity=Dashboard, fqdn=fqn)
        return EntityReference(id=dashboard.id, type="dashboard")


def get_table_key(row: Dict[str, Any]) -> Union[TableKey, None]:
    """
    Table key consists of schema and table name
    :param row:
    :return:
    """
    return TableKey(schema=row["schema"], table_name=row["table_name"])


class SampleDataSourceConfig(ConfigModel):
    sample_data_folder: str
    service_name: str = "bigquery_gcp"
    database: str = "warehouse"
    service_type: str = "BigQuery"
    scheme: str = "bigquery+pymysql"
    host_port: str = "9999"

    def get_sample_data_folder(self):
        return self.sample_data_folder

    def get_service_type(self):
        return self.service_type


@dataclass
class SampleDataSourceStatus(SourceStatus):
    success: List[str] = field(default_factory=list)
    failures: List[str] = field(default_factory=list)
    warnings: List[str] = field(default_factory=list)

    def scanned(self, entity_type: str, entity_name: str) -> None:
        self.success.append(entity_name)
        logger.info("{} Scanned: {}".format(entity_type, entity_name))

    def filtered(self, entity_type: str, entity_name: str, err: str) -> None:
        self.warnings.append(entity_name)
        logger.warning("Dropped {} {} due to {}".format(entity_type, entity_name, err))


class TableSchema:
    def __init__(self, filename):
        # error if the file is not csv file
        if not filename.endswith(".csv"):
            raise Exception("Input file should be a csv file")

        # file name is assumed to be the table name
        basename = os.path.basename(filename)
        self.table_name = os.path.splitext(basename)[0]

        with open(filename, "r") as fin:
            self.columns = [dict(i) for i in csv.DictReader(fin)]

    def primary_keys(self):
        return [c[COLUMN_NAME] for c in self.columns if c[KEY_TYPE] == "PK"]

    def foreign_keys(self):
        return [c[COLUMN_NAME] for c in self.columns if c[KEY_TYPE] == "FK"]

    def get_name(self):
        return self.table_name

    def get_schema(self):
        return self.columns

    def get_column_names(self):
        return [c[COLUMN_NAME] for c in self.columns]


class SampleTableMetadataGenerator:
    def __init__(self, table_to_df_dict, table_to_schema_map):
        self.table_to_df_dict = table_to_df_dict
        self.table_to_schema_map = table_to_schema_map
        self.sample_user = None
        self.sample_table = None
        self.sample_table_owner = None
        self.sample_table_last_updated = None

    def get_empty_dict_with_cols(self, columns):
        data = {}
        for c in columns:
            data[c] = []
        return data


class SampleDataSource(Source[Entity]):
    def __init__(
        self, config: SampleDataSourceConfig, metadata_config: MetadataServerConfig, ctx
    ):
        super().__init__(ctx)
        self.status = SampleDataSourceStatus()
        self.config = config
        self.metadata_config = metadata_config
        self.metadata = OpenMetadata(metadata_config)
        self.storage_service_json = json.load(
            open(self.config.sample_data_folder + "/locations/service.json", "r")
        )
        self.locations = json.load(
            open(self.config.sample_data_folder + "/locations/locations.json", "r")
        )
        self.storage_service = get_storage_service_or_create(
            self.storage_service_json,
            metadata_config,
        )
        self.glue_storage_service_json = json.load(
            open(self.config.sample_data_folder + "/glue/storage_service.json", "r")
        )
        self.glue_database_service_json = json.load(
            open(self.config.sample_data_folder + "/glue/database_service.json", "r")
        )
        self.glue_database = json.load(
            open(self.config.sample_data_folder + "/glue/database.json", "r")
        )
        self.glue_tables = json.load(
            open(self.config.sample_data_folder + "/glue/tables.json", "r")
        )
        self.glue_database_service = get_database_service_or_create_v2(
            self.glue_database_service_json,
            metadata_config,
        )
        self.glue_storage_service = get_storage_service_or_create(
            self.glue_storage_service_json,
            metadata_config,
        )
        self.database_service_json = json.load(
            open(self.config.sample_data_folder + "/datasets/service.json", "r")
        )
        self.database = json.load(
            open(self.config.sample_data_folder + "/datasets/database.json", "r")
        )
        self.tables = json.load(
            open(self.config.sample_data_folder + "/datasets/tables.json", "r")
        )
        self.database_service = get_database_service_or_create(
            config, self.metadata_config
        )
        self.kafka_service_json = json.load(
            open(self.config.sample_data_folder + "/topics/service.json", "r")
        )
        self.topics = json.load(
            open(self.config.sample_data_folder + "/topics/topics.json", "r")
        )
        self.kafka_service = get_messaging_service_or_create(
            self.kafka_service_json.get("name"),
            self.kafka_service_json.get("serviceType"),
            self.kafka_service_json.get("schemaRegistry"),
            self.kafka_service_json.get("brokers"),
            self.metadata_config,
        )
        self.dashboard_service_json = json.load(
            open(self.config.sample_data_folder + "/dashboards/service.json", "r")
        )
        self.charts = json.load(
            open(self.config.sample_data_folder + "/dashboards/charts.json", "r")
        )
        self.dashboards = json.load(
            open(self.config.sample_data_folder + "/dashboards/dashboards.json", "r")
        )
        self.dashboard_service = get_dashboard_service_or_create(
            self.dashboard_service_json.get("name"),
            self.dashboard_service_json.get("serviceType"),
            self.dashboard_service_json.get("username"),
            self.dashboard_service_json.get("password"),
            self.dashboard_service_json.get("dashboardUrl"),
            metadata_config,
        )
        self.pipeline_service_json = json.load(
            open(self.config.sample_data_folder + "/pipelines/service.json", "r")
        )
        self.pipelines = json.load(
            open(self.config.sample_data_folder + "/pipelines/pipelines.json", "r")
        )
        self.pipeline_service = get_pipeline_service_or_create(
            self.pipeline_service_json,
            metadata_config,
        )
        self.lineage = json.load(
            open(self.config.sample_data_folder + "/lineage/lineage.json", "r")
        )
        self.users = json.load(
            open(self.config.sample_data_folder + "/users/users.json", "r")
        )
        self.models = json.load(
            open(self.config.sample_data_folder + "/models/models.json", "r")
        )

    @classmethod
    def create(cls, config_dict, metadata_config_dict, ctx):
        config = SampleDataSourceConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(config, metadata_config, ctx)

    def prepare(self):
        pass

    def next_record(self) -> Iterable[Entity]:
        yield from self.ingest_locations()
        yield from self.ingest_glue()
        yield from self.ingest_tables()
        yield from self.ingest_topics()
        yield from self.ingest_charts()
        yield from self.ingest_dashboards()
        yield from self.ingest_pipelines()
        yield from self.ingest_lineage()
        yield from self.ingest_users()
        yield from self.ingest_mlmodels()

    def ingest_locations(self) -> Iterable[Location]:
        for location in self.locations["locations"]:
            location_ev = Location(
                id=uuid.uuid4(),
                name=location["name"],
                displayName=location["displayName"],
                description=location["description"],
                locationType=location["locationType"],
                service=EntityReference(
                    id=self.storage_service.id, type="storageService"
                ),
            )
            yield location_ev

    def ingest_glue(self) -> Iterable[OMetaDatabaseAndTable]:
        db = Database(
            id=uuid.uuid4(),
            name=self.glue_database["name"],
            description=self.glue_database["description"],
            service=EntityReference(
                id=self.glue_database_service.id,
                type=self.glue_database_service.serviceType.value,
            ),
        )
        for table in self.glue_tables["tables"]:
            table["id"] = uuid.uuid4()
            table_metadata = Table(**table)
            location_metadata = Location(
                id=uuid.uuid4(),
                name="s3://glue_bucket/dwh/schema/" + table["name"],
                description=table["description"],
                locationType=LocationType.Table,
                service=EntityReference(
                    id=self.glue_storage_service.id, type="storageService"
                ),
            )
            db_table_location = OMetaDatabaseAndTable(
                database=db, table=table_metadata, location=location_metadata
            )
            self.status.scanned("table", table_metadata.name.__root__)
            yield db_table_location

    def ingest_tables(self) -> Iterable[OMetaDatabaseAndTable]:
        db = Database(
            id=uuid.uuid4(),
            name=self.database["name"],
            description=self.database["description"],
            service=EntityReference(
                id=self.database_service.id, type=self.config.service_type
            ),
        )
        for table in self.tables["tables"]:
            table_metadata = Table(**table)
            table_and_db = OMetaDatabaseAndTable(table=table_metadata, database=db)
            self.status.scanned("table", table_metadata.name.__root__)
            yield table_and_db

    def ingest_topics(self) -> Iterable[CreateTopicEntityRequest]:
        for topic in self.topics["topics"]:
            topic["service"] = EntityReference(
                id=self.kafka_service.id, type="messagingService"
            )
            create_topic = CreateTopicEntityRequest(**topic)
            self.status.scanned("topic", create_topic.name.__root__)
            yield create_topic

    def ingest_charts(self) -> Iterable[Chart]:
        for chart in self.charts["charts"]:
            try:
                chart_ev = Chart(
                    name=chart["name"],
                    displayName=chart["displayName"],
                    description=chart["description"],
                    chart_type=chart["chartType"],
                    url=chart["chartUrl"],
                    service=EntityReference(
                        id=self.dashboard_service.id, type="dashboardService"
                    ),
                )
                self.status.scanned("chart", chart_ev.name)
                yield chart_ev
            except ValidationError as err:
                logger.error(err)

    def ingest_dashboards(self) -> Iterable[Dashboard]:
        for dashboard in self.dashboards["dashboards"]:
            dashboard_ev = Dashboard(
                name=dashboard["name"],
                displayName=dashboard["displayName"],
                description=dashboard["description"],
                url=dashboard["dashboardUrl"],
                charts=dashboard["charts"],
                service=EntityReference(
                    id=self.dashboard_service.id, type="dashboardService"
                ),
            )
            self.status.scanned("dashboard", dashboard_ev.name)
            yield dashboard_ev

    def ingest_pipelines(self) -> Iterable[Dashboard]:
        for pipeline in self.pipelines["pipelines"]:
            pipeline_ev = Pipeline(
                id=uuid.uuid4(),
                name=pipeline["name"],
                displayName=pipeline["displayName"],
                description=pipeline["description"],
                pipelineUrl=pipeline["pipelineUrl"],
                tasks=pipeline["tasks"],
                service=EntityReference(
                    id=self.pipeline_service.id, type="pipelineService"
                ),
            )
            yield pipeline_ev

    def ingest_lineage(self) -> Iterable[AddLineage]:
        for edge in self.lineage:
            from_entity_ref = get_lineage_entity_ref(edge["from"], self.metadata_config)
            to_entity_ref = get_lineage_entity_ref(edge["to"], self.metadata_config)
            lineage = AddLineage(
                edge=EntitiesEdge(fromEntity=from_entity_ref, toEntity=to_entity_ref)
            )
            yield lineage

    def ingest_mlmodels(self) -> Iterable[CreateMlModelEntityRequest]:
        """
        Convert sample model data into a Model Entity
        to feed the metastore
        """
        from metadata.generated.schema.entity.data.dashboard import Dashboard

        for model in self.models:
            # Fetch linked dashboard ID from name
            fqdn = model["dashboard"]
            dashboard = self.metadata.get_by_name(entity=Dashboard, fqdn=fqdn)

            if not dashboard:
                raise InvalidSampleDataException(
                    f"Cannot find {fqdn} in Sample Dashboards"
                )

            dashboard_id = str(dashboard.id.__root__)

            model_ev = CreateMlModelEntityRequest(
                name=model["name"],
                displayName=model["displayName"],
                description=model["description"],
                algorithm=model["algorithm"],
                dashboard=EntityReference(id=dashboard_id, type="dashboard"),
            )
            yield model_ev

    def ingest_users(self) -> Iterable[User]:
        try:
            for user in self.users["users"]:
                teams = [
                    EntityReference(
                        id=uuid.uuid4(),
                        name=user["teams"],
                        type="team",
                        description=f"This is {user['teams']} description.",
                    )
                ]
                user_metadata = User(
                    id=uuid.uuid4(),
                    name=user["name"],
                    displayName=user["displayName"],
                    email=user["email"],
                    teams=teams,
                    href=Href(__root__="http://localhost"),
                )
                yield user_metadata
        except Exception as err:
            logger.error(err)

    def close(self):
        pass

    def get_status(self):
        return self.status



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional

import pydruid

from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.ingestion.source.sql_source import SQLConnectionConfig, SQLSource


class DruidConfig(SQLConnectionConfig):
    scheme = "druid"
    auth_options: Optional[str] = None
    service_type = "Druid"

    def get_connection_url(self):
        url = super().get_connection_url()
        return f"{url}/druid/v2/sql"


class DruidSource(SQLSource):
    def __init__(self, config, metadata_config, ctx):
        super().__init__(config, metadata_config, ctx)

    @classmethod
    def create(cls, config_dict, metadata_config_dict, ctx):
        config = DruidConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(config, metadata_config, ctx)



# generated by datamodel-codegen:
#   filename:  schema/api/operations/workflows/createIngestion.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field, constr

from ....operations.workflows import ingestion
from ....type import basic, entityReference, tagLabel


class CreateIngestionEntityRequest(BaseModel):
    name: constr(min_length=1, max_length=128) = Field(
        ..., description='Name that identifies this ingestion instance uniquely.'
    )
    displayName: Optional[str] = Field(
        None, description='Display Name that identifies this Ingestion.'
    )
    description: Optional[str] = Field(None, description='Description of the workflow.')
    ingestionType: Optional[ingestion.IngestionType] = None
    owner: Optional[entityReference.EntityReference] = Field(
        None, description='Owner of this Ingestion.'
    )
    tags: Optional[List[tagLabel.TagLabel]] = Field(
        None, description='Tags associated with the Ingestion.'
    )
    forceDeploy: Optional[bool] = Field(
        'false',
        description='Deploy the workflow by overwriting existing workflow with the same name.',
    )
    pauseWorkflow: Optional[bool] = Field(
        'false',
        description='pause the workflow from running once the deploy is finished successfully.',
    )
    concurrency: Optional[int] = Field(1, description='Concurrency of the Pipeline.')
    startDate: basic.Date = Field(..., description='Start date of the workflow.')
    endDate: Optional[basic.Date] = Field(None, description='End Date of the workflow.')
    workflowTimezone: Optional[str] = Field(
        'UTC', description='Timezone in which workflow going to be scheduled.'
    )
    retries: Optional[int] = Field(1, description='Retry workflow in case of failure')
    retryDelay: Optional[int] = Field(
        300, description='Delay between retries in seconds.'
    )
    workflowCatchup: Optional[bool] = Field(
        'false', description='Workflow catchup for past executions.'
    )
    scheduleInterval: Optional[str] = Field(
        None, description='Scheduler Interval for the Workflow in cron format.'
    )
    workflowTimeout: Optional[int] = Field(
        60, description='Timeout for the workflow in seconds.'
    )
    service: entityReference.EntityReference = Field(
        ...,
        description='Link to the database service where this database is hosted in.',
    )
    connectorConfig: ingestion.ConnectorConfig



# generated by datamodel-codegen:
#   filename:  data/tags/personalDataTags.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import Any

from pydantic import BaseModel, Field


class Model(BaseModel):
    __root__: Any = Field(
        ...,
        description='Tags related classifying **Personal data** as defined by **GDPR.**<br/><br/>_Note to Legal - This tag category is provided as a starting point. Please review and update the tags based on your company policy. Also, add a reference to your GDPR policy document in this description._',
    )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import csv
import json
from datetime import datetime
from typing import Iterable

from metadata.ingestion.api.source import Source
from metadata.ingestion.models.table_queries import TableQuery
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.ingestion.source.sample_data import (
    SampleDataSourceConfig,
    SampleDataSourceStatus,
)
from metadata.utils.helpers import get_database_service_or_create


class SampleUsageSource(Source[TableQuery]):

    service_type = "BigQuery"

    def __init__(
        self, config: SampleDataSourceConfig, metadata_config: MetadataServerConfig, ctx
    ):
        super().__init__(ctx)
        self.status = SampleDataSourceStatus()
        self.config = config
        self.metadata_config = metadata_config
        self.service_json = json.load(
            open(config.sample_data_folder + "/datasets/service.json", "r")
        )
        self.query_log_csv = config.sample_data_folder + "/datasets/query_log"
        with open(self.query_log_csv, "r") as fin:
            self.query_logs = [dict(i) for i in csv.DictReader(fin)]
        self.service = get_database_service_or_create(self.config, metadata_config)

    @classmethod
    def create(cls, config_dict, metadata_config_dict, ctx):
        config = SampleDataSourceConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(config, metadata_config, ctx)

    def prepare(self):
        pass

    def next_record(self) -> Iterable[TableQuery]:
        for row in self.query_logs:
            tq = TableQuery(
                query=row["query"],
                user_name="",
                starttime="",
                endtime="",
                analysis_date=datetime.today().strftime("%Y-%m-%d %H:%M:%S"),
                database="shopify",
                aborted=False,
                sql=row["query"],
            )
            yield tq

    def close(self):
        pass

    def get_status(self):
        return self.status



# generated by datamodel-codegen:
#   filename:  schema/entity/policies/accessControl/rule.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Extra, Field

from .. import filters
from . import tagBased


class AccessControlRule(BaseModel):
    class Config:
        extra = Extra.forbid

    name: Optional[str] = Field(None, description='Name that identifies this Rule.')
    filters: filters.Filters1
    actions: List[tagBased.TagBased] = Field(
        ...,
        description='A set of access control enforcements to take on the entities.',
        min_items=1,
    )
    enabled: Optional[bool] = Field(True, description='Is the rule enabled.')



# generated by datamodel-codegen:
#   filename:  schema/type/changeEvent.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from enum import Enum
from typing import Any, List, Optional

from pydantic import BaseModel, Extra, Field

from . import basic, entityHistory


class EventType(Enum):
    entityCreated = 'entityCreated'
    entityUpdated = 'entityUpdated'
    entityDeleted = 'entityDeleted'


class EventFilter(BaseModel):
    eventType: EventType = Field(..., description='Event type that is being requested.')
    entities: Optional[List[str]] = Field(
        None,
        description='Entities for which the events are needed. Example - `table`, `topic`, etc. **When not set, events for all the entities will be provided**.',
    )


class ChangeEvent(BaseModel):
    class Config:
        extra = Extra.forbid

    eventType: EventType
    entityType: str = Field(
        ...,
        description='Entity type that changed. Use the schema of this entity to process the entity attribute.',
    )
    entityId: basic.Uuid = Field(
        ..., description='Identifier of entity that was modified by the operation.'
    )
    previousVersion: Optional[entityHistory.EntityVersion] = Field(
        None,
        description='Version of the entity before this change. Note that not all changes result in entity version change. When entity version is not changed, `previousVersion` is same as `currentVersion`.',
    )
    currentVersion: Optional[entityHistory.EntityVersion] = Field(
        None,
        description='Current version of the entity after this change. Note that not all changes result in entity version change. When entity version is not changed, `previousVersion` is same as `currentVersion`.',
    )
    userName: Optional[str] = Field(
        None, description='Name of the user whose activity resulted in the change.'
    )
    dateTime: basic.DateTime = Field(
        ..., description='Date and time when the change was made.'
    )
    changeDescription: Optional[entityHistory.ChangeDescription] = Field(
        None,
        description='For `eventType` `entityUpdated` this field captures details about what fields were added/updated/deleted. For `eventType` `entityCreated` or `entityDeleted` this field is null.',
    )
    entity: Optional[Any] = Field(
        None,
        description='For `eventType` `entityCreated`, this field captures JSON coded string of the entity using the schema corresponding to `entityType`.',
    )



# generated by datamodel-codegen:
#   filename:  schema/api/data/createMlModel.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field, constr

from ...entity.data import mlmodel
from ...type import basic, entityReference, tagLabel


class CreateMlModelEntityRequest(BaseModel):
    name: constr(min_length=1, max_length=128) = Field(
        ..., description='Name that identifies this ML model.'
    )
    displayName: Optional[str] = Field(
        None,
        description='Display Name that identifies this ML model. It could be title or label from the source services',
    )
    description: Optional[str] = Field(
        None,
        description='Description of the ML model instance. How it was trained and for what it is used.',
    )
    algorithm: str = Field(..., description='Algorithm used to train the ML Model')
    mlFeatures: Optional[List[mlmodel.MlFeature]] = Field(
        None, description='Features used to train the ML Model.'
    )
    mlHyperParameters: Optional[List[mlmodel.MlHyperParameter]] = Field(
        None, description='Hyper Parameters used to train the ML Model.'
    )
    dashboard: Optional[entityReference.EntityReference] = Field(
        None, description='Performance Dashboard URL to track metric evolution'
    )
    mlStore: Optional[mlmodel.MlStore] = Field(
        None,
        description='Location containing the ML Model. It can be a storage layer and/or a container repository.',
    )
    server: Optional[basic.Href] = Field(
        None,
        description='Endpoint that makes the ML Model available, e.g,. a REST API serving the data or computing predictions.',
    )
    tags: Optional[List[tagLabel.TagLabel]] = Field(
        None, description='Tags for this ML Model'
    )
    owner: Optional[entityReference.EntityReference] = Field(
        None, description='Owner of this database'
    )



# generated by datamodel-codegen:
#   filename:  schema/entity/data/metrics.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Extra, Field, constr

from ...type import basic, entityHistory, entityReference, tagLabel, usageDetails


class Metrics(BaseModel):
    class Config:
        extra = Extra.forbid

    id: basic.Uuid = Field(
        ..., description='Unique identifier that identifies this metrics instance.'
    )
    name: constr(min_length=1, max_length=128) = Field(
        ..., description='Name that identifies this metrics instance uniquely.'
    )
    fullyQualifiedName: Optional[constr(min_length=1, max_length=256)] = Field(
        None,
        description="A unique name that identifies a metric in the format 'ServiceName.MetricName'.",
    )
    displayName: Optional[str] = Field(
        None, description='Display Name that identifies this metric.'
    )
    description: Optional[str] = Field(
        None,
        description='Description of metrics instance, what it is, and how to use it.',
    )
    version: Optional[entityHistory.EntityVersion] = Field(
        None, description='Metadata version of the entity.'
    )
    updatedAt: Optional[basic.DateTime] = Field(
        None,
        description='Last update time corresponding to the new version of the entity.',
    )
    updatedBy: Optional[str] = Field(None, description='User who made the update.')
    href: Optional[basic.Href] = Field(
        None, description='Link to the resource corresponding to this entity.'
    )
    owner: Optional[entityReference.EntityReference] = Field(
        None, description='Owner of this metrics.'
    )
    tags: Optional[List[tagLabel.TagLabel]] = Field(
        None, description='Tags for this chart.'
    )
    service: entityReference.EntityReference = Field(
        ..., description='Link to service where this metrics is hosted in.'
    )
    usageSummary: Optional[usageDetails.TypeUsedToReturnUsageDetailsOfAnEntity] = Field(
        None, description='Latest usage information for this database.'
    )
    changeDescription: Optional[entityHistory.ChangeDescription] = Field(
        None, description='Change that lead to this version of the entity.'
    )



# generated by datamodel-codegen:
#   filename:  schema/api/feed/createThread.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from pydantic import BaseModel, Field

from ...type import basic


class CreateThreadEntityRequest(BaseModel):
    message: str = Field(..., description='Message')
    from_: basic.Uuid = Field(
        ...,
        alias='from',
        description='ID of User (regular user or bot) posting the message',
    )
    about: basic.EntityLink = Field(
        ...,
        description='Data asset about which this thread is created for with format <#E/{enties}/{entityName}/{field}/{fieldValue}',
    )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional
from urllib.parse import quote_plus

from pydantic import SecretStr

from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.ingestion.source.sql_source import SQLConnectionConfig, SQLSource


class AthenaConfig(SQLConnectionConfig):
    scheme: str = "awsathena+rest"
    username: Optional[str] = None
    password: Optional[SecretStr] = None
    database: Optional[str] = None
    aws_region: str
    s3_staging_dir: str
    work_group: str

    def get_connection_url(self):
        url = f"{self.scheme}://"
        if self.username:
            url += f"{quote_plus(self.username)}"
            if self.password:
                url += f":{quote_plus(self.password.get_secret_value())}"
        else:
            url += ":"
        url += f"@athena.{self.aws_region}.amazonaws.com:443/"
        if self.database:
            url += f"{self.database}"
        url += f"?s3_staging_dir={quote_plus(self.s3_staging_dir)}"
        url += f"&work_group={self.work_group}"

        return url


class AthenaSource(SQLSource):
    def __init__(self, config, metadata_config, ctx):
        super().__init__(config, metadata_config, ctx)

    @classmethod
    def create(cls, config_dict, metadata_config_dict, ctx):
        config = AthenaConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(config, metadata_config, ctx)



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import re
from dataclasses import dataclass
from typing import Any, Dict, Generic, List, Optional, TypeVar

from pydantic import BaseModel

T = TypeVar("T")

logger: logging.Logger = logging.getLogger(__name__)

# Allow types from the generated pydantic models
Entity = TypeVar("Entity", bound=BaseModel)


@dataclass
class WorkflowContext:
    workflow_id: str


class ConfigModel(BaseModel):
    class Config:
        extra = "forbid"


class DynamicTypedConfig(ConfigModel):
    type: str
    config: Optional[Any]


class WorkflowExecutionError(Exception):
    """An error occurred when executing the workflow"""


class IncludeFilterPattern(ConfigModel):
    """A class to store allow deny regexes"""

    includes: List[str] = [".*"]
    excludes: List[str] = []
    alphabet: str = "[A-Za-z0-9 _.-]"

    @property
    def alphabet_pattern(self):
        return re.compile(f"^{self.alphabet}+$")

    @classmethod
    def allow_all(cls):
        return IncludeFilterPattern()

    def included(self, string: str) -> bool:
        try:
            for exclude in self.excludes:
                if re.match(exclude, string):
                    return False

            for include in self.includes:
                if re.match(include, string):
                    return True
            return False
        except Exception as err:
            raise Exception("Regex Error: {}".format(err))

    def is_fully_specified_include_list(self) -> bool:
        for filter_pattern in self.includes:
            if not self.alphabet_pattern.match(filter_pattern):
                return False
        return True

    def get_allowed_list(self):
        assert self.is_fully_specified_include_list()
        return [a for a in self.includes if self.included(a)]



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import uuid
from dataclasses import dataclass, field
from typing import Dict, Iterable, List

import requests
from redash_toolbelt import Redash

from metadata.generated.schema.entity.data.chart import Chart
from metadata.generated.schema.entity.services.dashboardService import (
    DashboardServiceType,
)
from metadata.generated.schema.type.entityReference import EntityReference
from metadata.ingestion.api.common import ConfigModel, Entity, WorkflowContext
from metadata.ingestion.api.source import Source, SourceStatus
from metadata.ingestion.models.table_metadata import Chart as ModelChart
from metadata.ingestion.models.table_metadata import Dashboard
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.utils.helpers import get_dashboard_service_or_create


class RedashSourceConfig(ConfigModel):
    uri: str = "http://localhost:5000"
    username: str = ""
    api_key: str
    service_name: str
    service_type: str = "Redash"


@dataclass
class RedashSourceStatus(SourceStatus):
    items_scanned: int = 0
    filtered: List[str] = field(default_factory=list)

    def item_scanned_status(self) -> None:
        self.items_scanned += 1

    def item_dropped_status(self, item: str) -> None:
        self.filtered.append(item)


class RedashSource(Source[Entity]):
    config: RedashSourceConfig
    metadata_config: MetadataServerConfig
    status: RedashSourceStatus
    platform = "redash"
    dashboards_to_charts: Dict[str, List[str]]

    def __init__(
        self,
        config: RedashSourceConfig,
        metadata_config: MetadataServerConfig,
        ctx: WorkflowContext,
    ):
        super().__init__(ctx)
        self.config = config
        self.metadata_config = metadata_config
        self.status = RedashSourceStatus()
        self.client = Redash(self.config.uri, self.config.api_key)
        self.service = get_dashboard_service_or_create(
            config.service_name,
            DashboardServiceType.Redash.name,
            config.username,
            config.api_key,
            config.uri,
            metadata_config,
        )
        self.dashboards_to_charts = {}

    @classmethod
    def create(
        cls, config_dict: dict, metadata_config_dict: dict, ctx: WorkflowContext
    ):
        config = RedashSourceConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(config, metadata_config, ctx)

    def prepare(self):
        pass

    def next_record(self) -> Iterable[Entity]:
        yield from self.get_redash_charts()
        dashboard_info = self.client.dashboards()
        yield from self.get_redash_dashboard_charts(dashboard_info)
        yield from self.get_redash_dashboard(dashboard_info)

    def get_redash_charts(self) -> Chart:
        query_info = self.client.queries()
        for query_info in query_info["results"]:
            query_id = query_info["id"]
            query_name = query_info["name"]
            query_data = requests.get(
                f"{self.config.uri}/api/queries/{query_id}"
            ).json()
            for visualization in query_data.get("Visualizations", []):
                chart_type = visualization.get("type", "")
                chart_description = (
                    visualization.get("description", "")
                    if visualization.get("description", "")
                    else ""
                )
                yield Chart(
                    id=uuid.uuid4(),
                    name=query_id,
                    displayName=query_name,
                    chartType=chart_type,
                    service=EntityReference(
                        id=self.service.id, type="dashboardService"
                    ),
                    description=chart_description,
                )

    def get_redash_dashboard_charts(self, dashboard_info) -> Chart:
        for dashboard_info in dashboard_info["results"]:
            dashboard_id = dashboard_info["id"]
            if dashboard_id is not None:
                dashboard_data = self.client.dashboard(dashboard_info["slug"])
                self.dashboards_to_charts[dashboard_id] = []
                for widgets in dashboard_data.get("widgets", []):
                    visualization = widgets.get("visualization")
                    self.dashboards_to_charts[dashboard_id].append(widgets["id"])
                    yield ModelChart(
                        name=widgets["id"],
                        displayName=visualization["query"]["name"],
                        chart_type=visualization["type"],
                        service=EntityReference(
                            id=self.service.id, type="dashboardService"
                        ),
                        url=(
                            f"{self.config.uri}/dashboard/{dashboard_data.get('slug', '')}"
                        ),
                        description=visualization["description"],
                    )

    def get_redash_dashboard(self, dashboard_info) -> Dashboard:
        for dashboard_info in dashboard_info["results"]:
            dashboard_id = dashboard_info["id"]
            if dashboard_id is not None:
                self.status.item_scanned_status()
                dashboard_data = self.client.dashboard(dashboard_info["slug"])
                dashboard_url = (
                    f"{self.config.uri}/dashboard/{dashboard_data.get('slug', '')}"
                )
                for widgets in dashboard_data.get("widgets", []):
                    dashboard_description = widgets.get("text")
                yield Dashboard(
                    id=uuid.uuid4(),
                    name=dashboard_id,
                    displayName=dashboard_info["name"],
                    description=dashboard_description if dashboard_info else "",
                    charts=self.dashboards_to_charts[dashboard_id],
                    usageSummary=None,
                    service=EntityReference(
                        id=self.service.id, type="dashboardService"
                    ),
                    url=dashboard_url,
                )

    def get_status(self) -> SourceStatus:
        return self.status

    def close(self):
        self.client.session.close()



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from metadata.generated.schema.type.entityReference import EntityReference


class Table(BaseModel):
    """Table Fully Qualified Name"""

    fullyQualifiedName: str


class FieldChange(BaseModel):
    name: str
    newValue: Optional[str]
    oldValue: Optional[str]


class ChangeDescription(BaseModel):
    updatedBy: str
    updatedAt: int
    fieldsAdded: Optional[str]
    fieldsDeleted: Optional[str]
    fieldsUpdated: Optional[str]


class TableESDocument(BaseModel):
    """Elastic Search Mapping doc"""

    table_id: str
    database: str
    service: str
    service_type: str
    service_category: str
    entity_type: str = "table"
    name: str
    suggest: List[dict]
    description: Optional[str] = None
    table_type: Optional[str] = None
    last_updated_timestamp: Optional[int]
    column_names: List[str]
    column_descriptions: List[str]
    monthly_stats: int
    monthly_percentile_rank: int
    weekly_stats: int
    weekly_percentile_rank: int
    daily_stats: int
    daily_percentile_rank: int
    tags: List[str]
    fqdn: str
    tier: Optional[str] = None
    owner: str
    followers: List[str]
    change_descriptions: Optional[List[ChangeDescription]] = None
    doc_as_upsert: bool = True


class TopicESDocument(BaseModel):
    """Topic Elastic Search Mapping doc"""

    topic_id: str
    service: str
    service_type: str
    service_category: str
    entity_type: str = "topic"
    name: str
    suggest: List[dict]
    description: Optional[str] = None
    last_updated_timestamp: Optional[int]
    tags: List[str]
    fqdn: str
    tier: Optional[str] = None
    owner: str
    followers: List[str]
    change_descriptions: Optional[List[ChangeDescription]] = None
    doc_as_upsert: bool = True


class DashboardESDocument(BaseModel):
    """Elastic Search Mapping doc for Dashboards"""

    dashboard_id: str
    service: str
    service_type: str
    service_category: str
    entity_type: str = "dashboard"
    name: str
    suggest: List[dict]
    description: Optional[str] = None
    last_updated_timestamp: Optional[int]
    chart_names: List[str]
    chart_descriptions: List[str]
    tags: List[str]
    fqdn: str
    tier: Optional[str] = None
    owner: str
    followers: List[str]
    monthly_stats: int
    monthly_percentile_rank: int
    weekly_stats: int
    weekly_percentile_rank: int
    daily_stats: int
    daily_percentile_rank: int
    change_descriptions: Optional[List[ChangeDescription]] = None
    doc_as_upsert: bool = True


class PipelineESDocument(BaseModel):
    """Elastic Search Mapping doc for Pipelines"""

    pipeline_id: str
    service: str
    service_type: str
    service_category: str
    entity_type: str = "pipeline"
    name: str
    suggest: List[dict]
    description: Optional[str] = None
    last_updated_timestamp: Optional[int]
    task_names: List[str]
    task_descriptions: List[str]
    tags: List[str]
    fqdn: str
    tier: Optional[str] = None
    owner: str
    followers: List[str]
    change_descriptions: Optional[List[ChangeDescription]] = None
    doc_as_upsert: bool = True


class DbtModelESDocument(BaseModel):
    """Elastic Search Mapping doc"""

    dbt_model_id: str
    database: str
    service: str
    service_type: str
    service_category: str
    entity_type: str = "dbtmodel"
    name: str
    suggest: List[dict]
    description: Optional[str] = None
    dbt_type: Optional[str] = None
    last_updated_timestamp: Optional[int]
    column_names: List[str]
    column_descriptions: List[str]
    tags: List[str]
    fqdn: str
    tier: Optional[str] = None
    owner: str
    followers: List[str]
    change_descriptions: Optional[List[ChangeDescription]] = None
    doc_as_upsert: bool = True


class DashboardOwner(BaseModel):
    """Dashboard owner"""

    username: str
    first_name: str
    last_name: str


class Chart(BaseModel):
    """Chart"""

    name: str
    displayName: str
    description: str
    chart_type: str
    url: str
    owners: List[DashboardOwner] = None
    lastModified: int = None
    datasource_fqn: str = None
    service: EntityReference
    custom_props: Dict[Any, Any] = None


class Dashboard(BaseModel):
    """Dashboard"""

    name: str
    displayName: str
    description: str
    url: str
    owners: List[DashboardOwner] = None
    charts: List[str]
    service: EntityReference
    lastModified: int = None


class ValueFrequency(BaseModel):
    """Profiler ValueFrequency"""

    value: str
    frequency: int


class Histogram(BaseModel):
    """Histogram"""

    boundaries: List[str]
    heights: List[str]


class Quantile(BaseModel):
    """Quantile"""

    quantile: str
    value: str


class DatasetColumnProfile(BaseModel):
    """Dataset Column Profile stats"""

    fqdn: str
    unique_count: int = None
    unique_proportion: int = None
    null_count: int = None
    null_proportion: int = None
    min: str = None
    max: str = None
    mean: str = None
    median: str = None
    stddev: str = None
    quantiles: List[Quantile] = None
    distinct_value_frequencies: List[ValueFrequency] = None
    histogram: List[Histogram] = None
    sample_values: List[str] = None


class DatasetProfile(BaseModel):
    """Dataset(table) stats"""

    timestamp: int
    table_name: str
    row_count: int = None
    col_count: int = None
    col_profiles: List[DatasetColumnProfile] = None



# generated by datamodel-codegen:
#   filename:  schema/type/basic.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from datetime import date, datetime
from typing import Any, Optional
from uuid import UUID

from pydantic import AnyUrl, BaseModel, EmailStr, Field, constr


class Basic(BaseModel):
    __root__: Any = Field(
        ...,
        description='This schema defines basic common types that are used by other schemas.',
        title='Basic',
    )


class Uuid(BaseModel):
    __root__: UUID = Field(..., description='Unique id used to identify an entity.')


class Email(BaseModel):
    __root__: EmailStr = Field(
        ..., description='Email address of a user or other entities.'
    )


class Timestamp(BaseModel):
    __root__: str = Field(..., description='Timestamp in unixTimeMillis.')


class Href(BaseModel):
    __root__: AnyUrl = Field(..., description='URI that points to a resource.')


class TimeInterval(BaseModel):
    start: Optional[int] = Field(None, description='Start time in unixTimeMillis.')
    end: Optional[int] = Field(None, description='End time in unixTimeMillis.')


class Duration(BaseModel):
    __root__: str = Field(
        ..., description="Duration in ISO 8601 format in UTC. Example - 'P23DT23H'."
    )


class Date(BaseModel):
    __root__: date = Field(
        ..., description="Date in ISO 8601 format in UTC. Example - '2018-11-13'."
    )


class DateTime(BaseModel):
    __root__: datetime = Field(
        ...,
        description="Date and time in ISO 8601 format. Example - '2018-11-13T20:20:39+00:00'.",
    )


class EntityLink(BaseModel):
    __root__: constr(regex=r'^<#E/\S+/\S+>$') = Field(
        ...,
        description='Link to an entity or field within an entity using this format `<#E/{enties}/{entityName}/{field}/{fieldValue}`.',
    )


class SqlQuery(BaseModel):
    __root__: str = Field(
        ..., description="SQL query statement. Example - 'select * from orders'."
    )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import uuid
from dataclasses import dataclass, field
from typing import Iterable, List, Optional

from pydantic import SecretStr, ValidationError
from simple_salesforce import Salesforce

from metadata.generated.schema.entity.data.database import Database
from metadata.generated.schema.entity.data.table import (
    Column,
    Constraint,
    Table,
    TableData,
)
from metadata.generated.schema.type.entityReference import EntityReference
from metadata.ingestion.api.common import WorkflowContext
from metadata.ingestion.api.source import Source, SourceStatus
from metadata.ingestion.models.ometa_table_db import OMetaDatabaseAndTable
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.ingestion.source.sql_source import SQLConnectionConfig
from metadata.utils.helpers import get_database_service_or_create

logger: logging.Logger = logging.getLogger(__name__)


@dataclass
class SalesforceSourceStatus(SourceStatus):
    success: List[str] = field(default_factory=list)
    failures: List[str] = field(default_factory=list)
    warnings: List[str] = field(default_factory=list)
    filtered: List[str] = field(default_factory=list)

    def scanned(self, table_name: str) -> None:
        self.success.append(table_name)
        logger.info("Table Scanned: {}".format(table_name))

    def filter(
        self, table_name: str, err: str, dataset_name: str = None, col_type: str = None
    ) -> None:
        self.filtered.append(table_name)
        logger.warning("Dropped Table {} due to {}".format(table_name, err))


class SalesforceConfig(SQLConnectionConfig):
    username: str
    password: SecretStr
    security_token: str
    host_port: Optional[str]
    scheme: str
    service_type = "MySQL"
    sobject_name: str

    def get_connection_url(self):
        return super().get_connection_url()


class SalesforceSource(Source[OMetaDatabaseAndTable]):
    def __init__(
        self, config: SalesforceConfig, metadata_config: MetadataServerConfig, ctx
    ):
        super().__init__(ctx)
        self.config = config
        self.service = get_database_service_or_create(config, metadata_config)
        self.status = SalesforceSourceStatus()
        self.sf = Salesforce(
            username=self.config.username,
            password=self.config.password.get_secret_value(),
            security_token=self.config.security_token,
        )

    @classmethod
    def create(cls, config: dict, metadata_config: dict, ctx: WorkflowContext):
        config = SalesforceConfig.parse_obj(config)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config)
        return cls(config, metadata_config, ctx)

    def column_type(self, column_type: str):
        if column_type in ["ID", "PHONE", "CURRENCY"]:
            type = "INT"
        elif column_type in ["REFERENCE", "PICKLIST", "TEXTAREA", "ADDRESS", "URL"]:
            type = "VARCHAR"
        else:
            type = column_type
        return type

    def next_record(self) -> Iterable[OMetaDatabaseAndTable]:
        yield from self.salesforce_client()

    def fetch_sample_data(self, sobject_name):
        md = self.sf.restful("sobjects/{}/describe/".format(sobject_name), params=None)
        columns = []
        rows = []
        for column in md["fields"]:
            columns.append(column["name"])
        query = "select {} from {}".format(
            str(columns)[1:-1].replace("'", ""), sobject_name
        )
        logger.info("Ingesting data using {}".format(query))
        resp = self.sf.query(query)
        for record in resp["records"]:
            row = []
            for column in columns:
                row.append(record[f"{column}"])
            rows.append(row)
        return TableData(columns=columns, rows=rows)

    def salesforce_client(self) -> Iterable[OMetaDatabaseAndTable]:
        try:

            row_order = 1
            table_columns = []
            md = self.sf.restful(
                "sobjects/{}/describe/".format(self.config.sobject_name), params=None
            )

            for column in md["fields"]:
                col_constraint = None
                if column["nillable"]:
                    col_constraint = Constraint.NULL
                elif not column["nillable"]:
                    col_constraint = Constraint.NOT_NULL
                if column["unique"]:
                    col_constraint = Constraint.UNIQUE

                table_columns.append(
                    Column(
                        name=column["name"],
                        description=column["label"],
                        columnDataType=self.column_type(column["type"].upper()),
                        columnConstraint=col_constraint,
                        ordinalPosition=row_order,
                    )
                )
                row_order += 1
            table_data = self.fetch_sample_data(self.config.sobject_name)
            logger.info("Successfully Ingested the sample data")
            table_entity = Table(
                id=uuid.uuid4(),
                name=self.config.sobject_name,
                tableType="Regular",
                description=" ",
                columns=table_columns,
                sampleData=table_data,
            )
            self.status.scanned(f"{self.config.scheme}.{self.config.sobject_name}")
            database_entity = Database(
                name=self.config.scheme,
                service=EntityReference(
                    id=self.service.id, type=self.config.service_type
                ),
            )
            table_and_db = OMetaDatabaseAndTable(
                table=table_entity, database=database_entity
            )
            yield table_and_db
        except ValidationError as err:
            logger.error(err)
            self.status.failure("{}".format(self.config.sobject_name), err)

    def prepare(self):
        pass

    def get_status(self) -> SourceStatus:
        return self.status

    def close(self):
        pass



import re
from typing import Any, Dict, Optional, Set, Type

from sqlalchemy.sql import sqltypes as types

from metadata.ingestion.api.source import SourceStatus


def register_custom_type(tp: Type[types.TypeEngine], output: str = None) -> None:
    if output:
        _column_type_mapping[tp] = output
    else:
        _known_unknown_column_types.add(tp)


def register_custom_str_type(tp: str, output: str) -> None:
    _column_string_mapping[tp] = output


_column_type_mapping: Dict[Type[types.TypeEngine], str] = {
    types.Integer: "INT",
    types.Numeric: "INT",
    types.Boolean: "BOOLEAN",
    types.Enum: "ENUM",
    types._Binary: "BYTES",
    types.LargeBinary: "BYTES",
    types.PickleType: "BYTES",
    types.ARRAY: "ARRAY",
    types.VARCHAR: "VARCHAR",
    types.String: "STRING",
    types.Date: "DATE",
    types.DATE: "DATE",
    types.Time: "TIME",
    types.DateTime: "DATETIME",
    types.DATETIME: "DATETIME",
    types.TIMESTAMP: "TIMESTAMP",
    types.NullType: "NULL",
    types.JSON: "JSON",
    types.CHAR: "CHAR",
    types.DECIMAL: "DECIMAL",
    types.Interval: "INTERVAL",
}

_column_string_mapping = {
    "ARRAY": "ARRAY",
    "BIGINT": "BIGINT",
    "BIGNUMERIC": "NUMERIC",
    "BIGSERIAL": "BIGINT",
    "BINARY": "BINARY",
    "BIT": "INT",
    "BLOB": "BLOB",
    "BOOL": "BOOLEAN",
    "BOOLEAN": "BOOLEAN",
    "BPCHAR": "CHAR",
    "BYTEINT": "BYTEINT",
    "BYTES": "BYTES",
    "CHAR": "CHAR",
    "CHARACTER VARYING": "VARCHAR",
    "DATE": "DATE",
    "DATETIME": "DATETIME",
    "DATETIME2": "DATETIME",
    "DATETIMEOFFSET": "DATETIME",
    "DECIMAL": "DECIMAL",
    "DOUBLE PRECISION": "DOUBLE",
    "DOUBLE": "DOUBLE",
    "ENUM": "ENUM",
    "FLOAT4": "FLOAT",
    "FLOAT64": "DOUBLE",
    "FLOAT8": "DOUBLE",
    "GEOGRAPHY": "GEOGRAPHY",
    "HYPERLOGLOG": "BINARY",
    "IMAGE": "BINARY",
    "INT": "INT",
    "INT2": "SMALLINT",
    "INT4": "INT",
    "INT64": "BIGINT",
    "INT8": "BIGINT",
    "INTEGER": "INT",
    "INTERVAL DAY TO SECOND": "INTERVAL",
    "INTERVAL YEAR TO MONTH": "INTERVAL",
    "INTERVAL": "INTERVAL",
    "JSON": "JSON",
    "LONG RAW": "BINARY",
    "LONG VARCHAR": "VARCHAR",
    "LONGBLOB": "LONGBLOB",
    "MAP": "MAP",
    "MEDIUMBLOB": "MEDIUMBLOB",
    "MEDIUMINT": "INT",
    "MEDIUMTEXT": "MEDIUMTEXT",
    "MONEY": "NUMBER",
    "NCHAR": "CHAR",
    "NTEXT": "TEXT",
    "NULL": "NULL",
    "NUMBER": "NUMBER",
    "NUMERIC": "NUMBER",
    "NVARCHAR": "VARCHAR",
    "OBJECT": "JSON",
    "RAW": "BINARY",
    "REAL": "FLOAT",
    "ROWID": "VARCHAR",
    "ROWVERSION": "NUMBER",
    "SET": "SET",
    "SMALLDATETIME": "DATETIME",
    "SMALLINT": "SMALLINT",
    "SMALLMONEY": "NUMBER",
    "SMALLSERIAL": "SMALLINT",
    "SQL_VARIANT": "VARBINARY",
    "STRING": "STRING",
    "STRUCT": "STRUCT",
    "TABLE": "BINARY",
    "TEXT": "TEXT",
    "TIME": "TIME",
    "TIMESTAMP WITHOUT TIME ZONE": "TIMESTAMP",
    "TIMESTAMP": "TIMESTAMP",
    "TIMESTAMPTZ": "TIMESTAMP",
    "TIMETZ": "TIMESTAMP",
    "TINYINT": "TINYINT",
    "UNION": "UNION",
    "UROWID": "VARCHAR",
    "VARBINARY": "VARBINARY",
    "VARCHAR": "VARCHAR",
    "VARIANT": "JSON",
    "XML": "BINARY",
    "XMLTYPE": "BINARY",
    "CURSOR": "BINARY",
}

_known_unknown_column_types: Set[Type[types.TypeEngine]] = {
    types.Interval,
    types.CLOB,
}


def check_column_complex_type(
    status: SourceStatus, dataset_name: str, column_raw_type: Any, col_name: str
):
    arr_data_type = None
    col_obj = None
    data_type_display = None
    children = None
    if column_raw_type.startswith("struct<"):
        col_type = "STRUCT"
        col_obj = _handle_complex_data_types(
            status,
            dataset_name,
            f"{col_name}:{column_raw_type}",
        )
        if "children" in col_obj and col_obj["children"] is not None:
            children = col_obj["children"]
    elif column_raw_type.startswith("map<"):
        col_type = "MAP"
        col_obj = _handle_complex_data_types(
            status,
            dataset_name,
            f"{col_name}:{column_raw_type}",
        )
    elif column_raw_type.startswith("array<"):
        col_type = "ARRAY"
        arr_data_type = re.match(r"(?:array<)(\w*)(?:.*)", column_raw_type)
        arr_data_type = arr_data_type.groups()[0].upper()
    elif column_raw_type.startswith("uniontype<") or column_raw_type.startswith(
        "union<"
    ):
        col_type = "UNION"
    else:
        col_type = get_column_type(status, dataset_name, column_raw_type.split("(")[0])
        data_type_display = col_type
    if data_type_display is None:
        data_type_display = column_raw_type
    return col_type, data_type_display, arr_data_type, children


def get_column_type(status: SourceStatus, dataset_name: str, column_type: Any) -> str:
    type_class: Optional[str] = None
    for sql_type in _column_type_mapping.keys():
        if isinstance(column_type, sql_type):
            type_class = _column_type_mapping[sql_type]
            break
    if type_class is None or type_class == "NULL":
        for sql_type in _known_unknown_column_types:
            if isinstance(column_type, sql_type):
                type_class = "NULL"
                break
        for col_type in _column_string_mapping.keys():
            if str(column_type).split("(")[0].split("<")[0].upper() in col_type:
                type_class = _column_string_mapping.get(col_type)
                break
            else:
                type_class = None
    if type_class is None or type_class == "NULL":
        status.warning(
            dataset_name, f"unable to map type {column_type!r} to metadata schema"
        )
        type_class = "NULL"
    return type_class


def get_last_index(nested_str):
    counter = 1
    for index, i in enumerate(nested_str):
        if i == ">":
            counter -= 1
        elif i == "<":
            counter += 1
        if counter == 0:
            break
    index = index - counter
    return index


def get_array_type(col_type):
    col = {}
    col["dataType"] = "ARRAY"
    col_type = col_type[: get_last_index(col_type) + 2]
    col["dataTypeDisplay"] = col_type
    col["arrayDataType"] = (
        re.match(r"(?:array<)(\w*)(?:.*)", col_type).groups()[0].upper()
    )
    return col


def _handle_complex_data_types(status, dataset_name, raw_type: str, level=0):
    col = {}
    if re.match(r"([\w\s]*)(:)(.*)", raw_type):
        name, col_type = raw_type.lstrip("<").split(":", 1)
        col["name"] = name
    else:
        col["name"] = f"field_{level}"
        if raw_type.startswith("struct<"):
            col_type = raw_type
        else:
            col_type = raw_type.lstrip("<").split(":", 1)[0]
    if col_type.startswith("struct<"):
        children = []
        struct_type, col_type = re.match(r"(struct<)(.*)", col_type).groups()
        pluck_index = get_last_index(col_type)
        pluck_nested = col_type[: pluck_index + 1]
        col["dataTypeDisplay"] = struct_type + pluck_nested
        while pluck_nested != "":
            col["dataType"] = "STRUCT"
            plucked = col_type[: get_last_index(col_type)]
            counter = 0
            continue_next = False
            for index, type in enumerate(plucked.split(",")):
                if continue_next:
                    continue_next = False
                    continue
                if re.match(r"(\w*)(:)(struct)(.*)", type):
                    col_name, datatype, rest = re.match(
                        r"(\w*)(?::)(struct)(.*)", ",".join(plucked.split(",")[index:])
                    ).groups()
                    type = f"{col_name}:{datatype}{rest[:get_last_index(rest) + 2]}"
                elif type.startswith("struct"):
                    datatype, rest = re.match(
                        r"(struct)(.*)", ",".join(plucked.split(",")[index:])
                    ).groups()
                    type = f"{datatype}{rest[:get_last_index(rest) + 2]}"
                elif re.match(r"([\w\s]*)(:?)(map)(.*)", type):
                    get_map_type = ",".join(plucked.split(",")[index:])
                    type, col_type = re.match(
                        r"([\w]*:?map<[\w,]*>)(.*)", get_map_type
                    ).groups()
                    continue_next = True
                elif re.match(r"([\w\s]*)(:?)(uniontype)(.*)", type):
                    get_union_type = ",".join(plucked.split(",")[index:])
                    type, col_type = re.match(
                        r"([\w\s]*:?uniontype<[\w\s,]*>)(.*)", get_union_type
                    ).groups()
                    continue_next = True
                children.append(
                    _handle_complex_data_types(status, dataset_name, type, counter)
                )
                if plucked.endswith(type):
                    break
                counter += 1
            pluck_nested = col_type[get_last_index(col_type) + 3 :]
            col["children"] = children
    elif col_type.startswith("array"):
        col.update(get_array_type(col_type))
    elif col_type.startswith("map"):
        col["dataType"] = "MAP"
        col["dataTypeDisplay"] = col_type
    elif col_type.startswith("uniontype"):
        col["dataType"] = "UNION"
        col["dataTypeDisplay"] = col_type
    else:
        if re.match(r"(?:[\w\s]*)(?:\()([\d]*)(?:\))", col_type):
            col["dataLength"] = re.match(
                r"(?:[\w\s]*)(?:\()([\d]*)(?:\))", col_type
            ).groups()[0]
        else:
            col["dataLength"] = 1
        col["dataType"] = get_column_type(
            status,
            dataset_name,
            re.match(r"([\w\s]*)(?:.*)", col_type).groups()[0],
        )
        col["dataTypeDisplay"] = col_type.rstrip(">")
    return col



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import importlib
import logging
from typing import Any, Iterable, Iterator, Optional, Union

import neo4j
from neo4j import GraphDatabase

from metadata.config.common import ConfigModel

logger: logging.Logger = logging.getLogger(__name__)


class Neo4JConfig(ConfigModel):
    username: Optional[str] = None
    password: Optional[str] = None
    neo4j_url: str
    max_connection_life_time: int = 50
    neo4j_encrypted: bool = True
    neo4j_validate_ssl: bool = False
    model_class: str = None


class Neo4jHelper:
    """
    A helper class to extract data from Neo4J
    """

    def __init__(self, conf: Neo4JConfig) -> None:
        """
        Establish connections and import data model class if provided
        :param conf:
        """
        self.conf = conf
        self.graph_url = self.conf.neo4j_url
        self.driver = self._get_driver()
        self._extract_iter: Union[None, Iterator] = None

        model_class = self.conf.model_class
        if model_class is not None:
            module_name, class_name = model_class.rsplit(".", 1)
            mod = importlib.import_module(module_name)
            self.model_class = getattr(mod, class_name)

    def _get_driver(self) -> Any:
        """
        Create a Neo4j connection to Database
        """
        trust = (
            neo4j.TRUST_SYSTEM_CA_SIGNED_CERTIFICATES
            if self.conf.neo4j_validate_ssl
            else neo4j.TRUST_ALL_CERTIFICATES
        )
        return GraphDatabase.driver(
            self.graph_url,
            auth=(self.conf.username, self.conf.password),
            encrypted=self.conf.neo4j_encrypted,
            trust=trust,
        )

    def _execute_query(self, tx: Any, query: str) -> Any:
        """
        Create an iterator to execute sql.
        """
        logger.debug("Executing query %s", query)
        result = tx.run(query)
        entities = []
        for record in result:
            entities.append(record.data())
        return entities

    def execute_query(self, query: str) -> Iterable[Any]:
        """
        Execute {query} and yield result one at a time
        """
        with self.driver.session() as session:
            neo4j_results = session.read_transaction(self._execute_query, query)
            if hasattr(self, "model_class"):
                results = [
                    self.model_class(**neo4j_result) for neo4j_result in neo4j_results
                ]
            else:
                results = neo4j_results
            return iter(results)

    def close(self) -> None:
        """
        close connection to neo4j cluster
        """
        try:
            self.driver.close()
        except Exception as e:
            logger.error("Exception encountered while closing the graph driver", e)



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import ast
import logging
from dataclasses import dataclass, field
from typing import Iterable, List, Optional

from mlflow.entities import RunData
from mlflow.entities.model_registry import ModelVersion
from mlflow.tracking import MlflowClient

from metadata.generated.schema.api.data.createMlModel import CreateMlModelEntityRequest
from metadata.generated.schema.entity.data.mlmodel import (
    FeatureType,
    MlFeature,
    MlHyperParameter,
    MlStore,
)
from metadata.ingestion.api.common import ConfigModel, WorkflowContext
from metadata.ingestion.api.source import Source, SourceStatus

logger: logging.Logger = logging.getLogger(__name__)


@dataclass
class MlFlowStatus(SourceStatus):
    """
    ML Model specific Status
    """

    success: List[str] = field(default_factory=list)
    failures: List[str] = field(default_factory=list)
    warnings: List[str] = field(default_factory=list)

    def scanned(self, model_name: str) -> None:
        """
        Log successful ML Model scans
        """
        self.success.append(model_name)
        logger.info(f"ML Model scanned: {model_name}")

    def failed(self, model_name: str, reason: str) -> None:
        """
        Log failed ML Model scans
        """
        self.failures.append(model_name)
        logger.error(f"ML Model failed: {model_name} - {reason}")

    def warned(self, model_name: str, reason: str) -> None:
        """
        Log Ml Model with warnings
        """
        self.warnings.append(model_name)
        logger.warning(f"ML Model warning: {model_name} - {reason}")


class MlFlowConnectionConfig(ConfigModel):
    """
    Required information to extract data from MLFlow
    """

    tracking_uri: str
    registry_uri: Optional[str]


class MlflowSource(Source[CreateMlModelEntityRequest]):
    """
    Source implementation to ingest MLFlow data.

    We will iterate on the registered ML Models
    and prepare an iterator of CreateMlModelEntityRequest
    """

    def __init__(self, config: MlFlowConnectionConfig, ctx: WorkflowContext):
        super().__init__(ctx)
        self.client = MlflowClient(
            tracking_uri=config.tracking_uri,
            registry_uri=config.registry_uri if config.registry_uri else None,
        )
        self.status = MlFlowStatus()

    def prepare(self):
        pass

    @classmethod
    def create(
        cls, config_dict: dict, metadata_config_dict: dict, ctx: WorkflowContext
    ):
        config = MlFlowConnectionConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def next_record(self) -> Iterable[CreateMlModelEntityRequest]:
        """
        Fetch all registered models from MlFlow.

        We are setting the `algorithm` to a constant
        as there is not a non-trivial generic approach
        for retrieving the algorithm from the registry.
        """
        for model in self.client.list_registered_models():

            # Get the latest version
            latest_version: Optional[ModelVersion] = next(
                (
                    ver
                    for ver in model.latest_versions
                    if ver.last_updated_timestamp == model.last_updated_timestamp
                ),
                None,
            )
            if not latest_version:
                self.status.failed(model.name, reason="Invalid version")
                continue

            run = self.client.get_run(latest_version.run_id)

            self.status.scanned(model.name)

            yield CreateMlModelEntityRequest(
                name=model.name,
                description=model.description,
                algorithm="mlflow",  # Setting this to a constant
                mlHyperParameters=self._get_hyper_params(run.data),
                mlFeatures=self._get_ml_features(
                    run.data, latest_version.run_id, model.name
                ),
                mlStore=self._get_ml_store(latest_version),
            )

    @staticmethod
    def _get_hyper_params(data: RunData) -> Optional[List[MlHyperParameter]]:
        """
        Get the hyper parameters from the parameters
        logged in the run data object.
        """
        if data.params:
            return [
                MlHyperParameter(name=param[0], value=param[1])
                for param in data.params.items()
            ]

        return None

    @staticmethod
    def _get_ml_store(version: ModelVersion) -> Optional[MlStore]:
        """
        Get the Ml Store from the model version object
        """
        if version.source:
            return MlStore(storage=version.source)
        return None

    def _get_ml_features(
        self, data: RunData, run_id: str, model_name: str
    ) -> Optional[List[MlFeature]]:
        """
        The RunData object comes with stringified `tags`.
        Let's transform those and try to extract the `signature`
        information
        """
        if data.tags:
            try:
                props = ast.literal_eval(data.tags["mlflow.log-model.history"])
                latest_props = next(
                    (prop for prop in props if prop["run_id"] == run_id), None
                )
                if not latest_props:
                    reason = f"Cannot find the run ID properties for {run_id}"
                    logging.warning(reason)
                    self.status.warned(model_name, reason)
                    return None

                if latest_props.get("signature") and latest_props["signature"].get(
                    "inputs"
                ):

                    features = ast.literal_eval(latest_props["signature"]["inputs"])

                    return [
                        MlFeature(
                            name=feature["name"],
                            dataType=FeatureType.categorical
                            if feature["type"] == "string"
                            else FeatureType.numerical,
                        )
                        for feature in features
                    ]

            except Exception as exc:
                reason = f"Cannot extract properties from RunData {exc}"
                logging.warning(reason)
                self.status.warned(model_name, reason)

        return None

    def get_status(self) -> SourceStatus:
        return self.status

    def close(self) -> None:
        """
        Don't need to close the client
        """
        pass



from django.contrib import admin
from django.urls import path

from .views import detail_page, list_page

urlpatterns = [
    path("admin/", admin.site.urls),
    path("", list_page),
    path("dataset/<str:fqn>/", detail_page, name="dataset"),
]
handler404 = "metadata_server.openmetadata.views.page_not_found_view"



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import importlib
import logging
import uuid
from typing import Optional, Type, TypeVar

import click
from pydantic import Field

from metadata.config.common import (
    ConfigModel,
    DynamicTypedConfig,
    WorkflowExecutionError,
)
from metadata.ingestion.api.bulk_sink import BulkSink
from metadata.ingestion.api.common import WorkflowContext
from metadata.ingestion.api.processor import Processor
from metadata.ingestion.api.sink import Sink
from metadata.ingestion.api.source import Source
from metadata.ingestion.api.stage import Stage

logger = logging.getLogger(__name__)

T = TypeVar("T")


class WorkflowConfig(ConfigModel):
    run_id: str = Field(default_factory=lambda: str(uuid.uuid1()))
    source: DynamicTypedConfig
    metadata_server: DynamicTypedConfig
    processor: Optional[DynamicTypedConfig] = None
    sink: Optional[DynamicTypedConfig] = None
    stage: Optional[DynamicTypedConfig] = None
    bulk_sink: Optional[DynamicTypedConfig] = None


class Workflow:
    config: WorkflowConfig
    ctx: WorkflowContext
    source: Source
    processor: Processor
    stage: Stage
    sink: Sink
    bulk_sink: BulkSink
    report = {}

    def __init__(self, config: WorkflowConfig):
        self.config = config
        self.ctx = WorkflowContext(workflow_id=self.config.run_id)
        source_type = self.config.source.type
        source_class = self.get(
            "metadata.ingestion.source.{}.{}Source".format(
                self.typeClassFetch(source_type, True),
                self.typeClassFetch(source_type, False),
            )
        )
        metadata_config = self.config.metadata_server.dict().get("config", {})
        self.source: Source = source_class.create(
            self.config.source.dict().get("config", {}), metadata_config, self.ctx
        )
        logger.debug(f"Source type:{source_type},{source_class} configured")
        self.source.prepare()
        logger.debug(f"Source type:{source_type},{source_class}  prepared")

        if self.config.processor:
            processor_type = self.config.processor.type
            processor_class = self.get(
                "metadata.ingestion.processor.{}.{}Processor".format(
                    self.typeClassFetch(processor_type, True),
                    self.typeClassFetch(processor_type, False),
                )
            )
            processor_config = self.config.processor.dict().get("config", {})
            self.processor: Processor = processor_class.create(
                processor_config, metadata_config, self.ctx
            )
            logger.debug(
                f"Processor Type: {processor_type}, {processor_class} configured"
            )

        if self.config.stage:
            stage_type = self.config.stage.type
            stage_class = self.get(
                "metadata.ingestion.stage.{}.{}Stage".format(
                    self.typeClassFetch(stage_type, True),
                    self.typeClassFetch(stage_type, False),
                )
            )
            stage_config = self.config.stage.dict().get("config", {})
            self.stage: Stage = stage_class.create(
                stage_config, metadata_config, self.ctx
            )
            logger.debug(f"Stage Type: {stage_type}, {stage_class} configured")

        if self.config.sink:
            sink_type = self.config.sink.type
            sink_class = self.get(
                "metadata.ingestion.sink.{}.{}Sink".format(
                    self.typeClassFetch(sink_type, True),
                    self.typeClassFetch(sink_type, False),
                )
            )
            sink_config = self.config.sink.dict().get("config", {})
            self.sink: Sink = sink_class.create(sink_config, metadata_config, self.ctx)
            logger.debug(f"Sink type:{self.config.sink.type},{sink_class} configured")

        if self.config.bulk_sink:
            bulk_sink_type = self.config.bulk_sink.type
            bulk_sink_class = self.get(
                "metadata.ingestion.bulksink.{}.{}BulkSink".format(
                    self.typeClassFetch(bulk_sink_type, True),
                    self.typeClassFetch(bulk_sink_type, False),
                )
            )
            bulk_sink_config = self.config.bulk_sink.dict().get("config", {})
            self.bulk_sink: BulkSink = bulk_sink_class.create(
                bulk_sink_config, metadata_config, self.ctx
            )
            logger.info(
                f"BulkSink type:{self.config.bulk_sink.type},{bulk_sink_class} configured"
            )

    def typeClassFetch(self, type: str, isFile: bool):
        if isFile:
            return type.replace("-", "_")
        else:
            return "".join([i.title() for i in type.replace("-", "_").split("_")])

    def get(self, key: str) -> Type[T]:
        if key.find(".") >= 0:
            # If the key contains a dot, we treat it as a import path and attempt
            # to load it dynamically.
            module_name, class_name = key.rsplit(".", 1)
            MyClass = getattr(importlib.import_module(module_name), class_name)
            return MyClass

    @classmethod
    def create(cls, config_dict: dict) -> "Workflow":
        config = WorkflowConfig.parse_obj(config_dict)
        return cls(config)

    def execute(self):
        for record in self.source.next_record():
            self.report["Source"] = self.source.get_status().as_obj()
            if hasattr(self, "processor"):
                processed_record = self.processor.process(record)
            else:
                processed_record = record
            if hasattr(self, "stage"):
                self.stage.stage_record(processed_record)
                self.report["Stage"] = self.stage.get_status().as_obj()
            if hasattr(self, "sink"):
                self.sink.write_record(processed_record)
                self.report["sink"] = self.sink.get_status().as_obj()

    def stop(self):
        if hasattr(self, "processor"):
            self.processor.close()
        if hasattr(self, "stage"):
            self.stage.close()
        if hasattr(self, "bulk_sink"):
            self.bulk_sink.write_records()
            self.report["Bulk_Sink"] = self.bulk_sink.get_status().as_obj()
            self.bulk_sink.close()
        if hasattr(self, "sink"):
            self.sink.close()
        self.source.close()

    def raise_from_status(self, raise_warnings=False):
        if self.source.get_status().failures:
            raise WorkflowExecutionError(
                "Source reported errors", self.source.get_status()
            )
        if hasattr(self, "sink") and self.sink.get_status().failures:
            raise WorkflowExecutionError("Sink reported errors", self.sink.get_status())
        if hasattr(self, "stage") and self.stage.get_status().failures:
            raise WorkflowExecutionError(
                "stage reported errors", self.stage.get_status()
            )
        if hasattr(self, "bulk_sink") and self.bulk_sink.get_status().failures:
            raise WorkflowExecutionError(
                "Bulk Sink reported errors", self.bulk_sink.get_status()
            )
        if raise_warnings and (
            self.source.get_status().warnings or self.sink.get_status().warnings
        ):
            raise WorkflowExecutionError(
                "Source reported warnings", self.source.get_status()
            )

    def print_status(self) -> int:
        click.echo()
        click.secho("Source Status:", bold=True)
        click.echo(self.source.get_status().as_string())
        if hasattr(self, "stage"):
            click.secho("Stage Status:", bold=True)
            click.echo(self.stage.get_status().as_string())
            click.echo()
        if hasattr(self, "sink"):
            click.secho("Sink Status:", bold=True)
            click.echo(self.sink.get_status().as_string())
            click.echo()
        if hasattr(self, "bulk_sink"):
            click.secho("Bulk Sink Status:", bold=True)
            click.echo(self.bulk_sink.get_status().as_string())
            click.echo()

        if self.source.get_status().failures or (
            hasattr(self, "sink") and self.sink.get_status().failures
        ):
            click.secho("Workflow finished with failures", fg="bright_red", bold=True)
            return 1
        elif self.source.get_status().warnings or (
            hasattr(self, "sink") and self.sink.get_status().warnings
        ):
            click.secho("Workflow finished with warnings", fg="yellow", bold=True)
            return 0
        else:
            click.secho("Workflow finished successfully", fg="green", bold=True)
            return 0



# generated by datamodel-codegen:
#   filename:  schema/api/services/createMessagingService.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import Optional

from pydantic import AnyUrl, BaseModel, Field, constr

from ...entity.services import messagingService
from ...type import schedule


class CreateMessagingServiceEntityRequest(BaseModel):
    name: constr(min_length=1, max_length=128) = Field(
        ..., description='Name that identifies the this entity instance uniquely'
    )
    description: Optional[str] = Field(
        None, description='Description of messaging service entity.'
    )
    serviceType: messagingService.MessagingServiceType
    brokers: messagingService.Brokers = Field(
        ...,
        description='Multiple bootstrap addresses for Kafka. Single proxy address for Pulsar.',
    )
    schemaRegistry: Optional[AnyUrl] = Field(None, description='Schema registry URL')
    ingestionSchedule: Optional[schedule.Schedule] = Field(
        None, description='Schedule for running metadata ingestion jobs'
    )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

CLUSTER_NODE_LABEL = "Cluster"

CLUSTER_RELATION_TYPE = "CLUSTER"
CLUSTER_REVERSE_RELATION_TYPE = "CLUSTER_OF"

CLUSTER_NAME_PROP_KEY = "name"



# generated by datamodel-codegen:
#   filename:  schema/entity/data/topic.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Extra, Field, conint, constr

from ...type import basic, entityHistory, entityReference, tagLabel
from ..services import messagingService


class TopicName(BaseModel):
    __root__: constr(regex=r'^[^.]*$', min_length=1, max_length=128) = Field(
        ..., description='Name that identifies a topic.'
    )


class SchemaType(Enum):
    Avro = 'Avro'
    Protobuf = 'Protobuf'
    JSON = 'JSON'
    Other = 'Other'


class CleanupPolicy(Enum):
    delete = 'delete'
    compact = 'compact'


class Topic(BaseModel):
    class Config:
        extra = Extra.forbid

    id: basic.Uuid = Field(
        ..., description='Unique identifier that identifies this topic instance.'
    )
    name: TopicName = Field(..., description='Name that identifies the topic.')
    fullyQualifiedName: Optional[str] = Field(
        None,
        description="Name that uniquely identifies a topic in the format 'messagingServiceName.topicName'.",
    )
    displayName: Optional[str] = Field(
        None,
        description='Display Name that identifies this topic. It could be title or label from the source services.',
    )
    description: Optional[str] = Field(
        None, description='Description of the topic instance.'
    )
    version: Optional[entityHistory.EntityVersion] = Field(
        None, description='Metadata version of the entity.'
    )
    updatedAt: Optional[basic.DateTime] = Field(
        None,
        description='Last update time corresponding to the new version of the entity.',
    )
    updatedBy: Optional[str] = Field(None, description='User who made the update.')
    service: entityReference.EntityReference = Field(
        ...,
        description='Link to the messaging cluster/service where this topic is hosted in.',
    )
    serviceType: Optional[messagingService.MessagingServiceType] = Field(
        None, description='Service type where this topic is hosted in.'
    )
    partitions: conint(ge=1) = Field(
        ..., description='Number of partitions into which the topic is divided.'
    )
    schemaText: Optional[str] = Field(
        None,
        description='Schema used for message serialization. Optional as some topics may not have associated schemas.',
    )
    schemaType: Optional[SchemaType] = Field(
        None, description='Schema used for message serialization.'
    )
    cleanupPolicies: Optional[List[CleanupPolicy]] = Field(
        None,
        description='Topic clean up policies. For Kafka - `cleanup.policy` configuration.',
    )
    retentionTime: Optional[float] = Field(
        None,
        description='Retention time in milliseconds. For Kafka - `retention.ms` configuration.',
    )
    replicationFactor: Optional[int] = Field(
        None, description='Replication Factor in integer (more than 1).'
    )
    maximumMessageSize: Optional[int] = Field(
        None,
        description='Maximum message size in bytes. For Kafka - `max.message.bytes` configuration.',
    )
    minimumInSyncReplicas: Optional[int] = Field(
        None,
        description='Minimum number replicas in sync to control durability. For Kafka - `min.insync.replicas` configuration.',
    )
    retentionSize: Optional[float] = Field(
        '-1',
        description='Maximum size of a partition in bytes before old data is discarded. For Kafka - `retention.bytes` configuration.',
    )
    owner: Optional[entityReference.EntityReference] = Field(
        None, description='Owner of this topic.'
    )
    followers: Optional[entityReference.EntityReferenceList] = Field(
        None, description='Followers of this table.'
    )
    tags: Optional[List[tagLabel.TagLabel]] = Field(
        None, description='Tags for this table.'
    )
    href: Optional[basic.Href] = Field(
        None, description='Link to the resource corresponding to this entity.'
    )
    changeDescription: Optional[entityHistory.ChangeDescription] = Field(
        None, description='Change that lead to this version of the entity.'
    )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from typing import Dict, Set

from setuptools import find_namespace_packages, setup


def get_version():
    root = os.path.dirname(__file__)
    changelog = os.path.join(root, "CHANGELOG")
    with open(changelog) as f:
        return f.readline().strip()


def get_long_description():
    root = os.path.dirname(__file__)
    with open(os.path.join(root, "README.md")) as f:
        description = f.read()
    description += "\n\nChangelog\n=========\n\n"
    with open(os.path.join(root, "CHANGELOG")) as f:
        description += f.read()
    return description


base_requirements = {
    "commonregex",
    "idna<3,>=2.5",
    "click>=7.1.1",
    "expandvars>=0.6.5"
    "dataclasses>=0.8"
    "typing_extensions>=3.7.4"
    "mypy_extensions>=0.4.3",
    "typing-inspect",
    "pydantic>=1.7.4",
    "pydantic[email]>=1.7.2",
    "google>=3.0.0",
    "google-auth>=1.33.0",
    "python-dateutil>=2.8.1",
    "email-validator>=1.0.3",
    "wheel~=0.36.2",
    "python-jose==3.3.0",
    "sqlalchemy>=1.3.24",
    "sql-metadata~=2.0.0",
    "requests~=2.26",
    "cryptography",
    "PyYAML",
}

report_requirements = {
    "asgiref==3.4.1",
    "Django==3.2.7",
    "pytz==2021.1",
    "sqlparse==0.4.2",
}

base_plugins = {
    "query-parser",
    "metadata-usage",
    "file-stage",
    "sql-metadata~=2.0.0",
}
plugins: Dict[str, Set[str]] = {
    "amundsen": {"neo4j~=4.4.0"},
    "athena": {"PyAthena[SQLAlchemy]"},
    "bigquery": {"openmetadata-sqlalchemy-bigquery==0.2.2"},
    "bigquery-usage": {"google-cloud-logging", "cachetools"},
    "docker": {"docker==5.0.3"},
    "dbt": {},
    "druid": {"pydruid>=0.6.2"},
    "elasticsearch": {"elasticsearch~=7.13.1"},
    "glue": {"boto3~=1.19.12"},
    "hive": {
        "openmetadata-sqlalchemy-hive==0.2.0",
        "thrift~=0.13.0",
        "sasl==0.3.1",
        "thrift-sasl==0.4.3",
    },
    "kafka": {"confluent_kafka>=1.5.0", "fastavro>=1.2.0"},
    "ldap-users": {"ldap3==2.9.1"},
    "looker": {"looker-sdk==21.12.2"},
    "mssql": {"sqlalchemy-pytds>=0.3"},
    "mssql-odbc": {"pyodbc"},
    "mysql": {"pymysql>=1.0.2"},
    "oracle": {"cx_Oracle"},
    "presto": {"pyhive~=0.6.3"},
    "trino": {"sqlalchemy-trino"},
    "postgres": {"pymysql>=1.0.2", "psycopg2-binary", "GeoAlchemy2"},
    "redash": {"redash-toolbelt==0.1.4"},
    "redshift": {
        "openmetadata-sqlalchemy-redshift==0.2.1",
        "psycopg2-binary",
        "GeoAlchemy2",
    },
    "redshift-usage": {
        "openmetadata-sqlalchemy-redshift==0.2.1",
        "psycopg2-binary",
        "GeoAlchemy2",
    },
    "data-profiler": {"openmetadata-data-profiler"},
    "snowflake": {"snowflake-sqlalchemy<=1.3.2"},
    "snowflake-usage": {"snowflake-sqlalchemy<=1.3.2"},
    "sample-entity": {"faker~=8.1.1"},
    "superset": {},
    "tableau": {"tableau-api-lib==0.1.22"},
    "vertica": {"sqlalchemy-vertica[vertica-python]>=0.0.5"},
    "report-server": report_requirements,
    "airflow": {"apache-airflow >= 1.10.2"},
    "salesforce": {"simple_salesforce~=1.11.4"},
    "okta": {"okta~=2.3.0"},
    "mlflow": {"mlflow-skinny~=1.22.0"},
}

build_options = {"includes": ["_cffi_backend"]}
setup(
    name="openmetadata-ingestion",
    version="0.7.3",
    url="https://open-metadata.org/",
    author="OpenMetadata Committers",
    license="Apache License 2.0",
    description="Ingestion Framework for OpenMetadata",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    python_requires=">=3.8",
    options={"build_exe": build_options},
    package_dir={"": "src"},
    zip_safe=False,
    dependency_links=[],
    project_urls={
        "Documentation": "https://docs.open-metadata.org/",
        "Source": "https://github.com/open-metadata/OpenMetadata",
    },
    packages=find_namespace_packages(where="./src", exclude=["tests*"]),
    entry_points={
        "console_scripts": ["metadata = metadata.cmd:metadata"],
        "apache_airflow_provider": [
            "provider_info = airflow_provider_openmetadata:get_provider_config"
        ],
    },
    install_requires=list(base_requirements),
    extras_require={
        "base": list(base_requirements),
        **{plugin: list(dependencies) for (plugin, dependencies) in plugins.items()},
        "all": list(
            base_requirements.union(
                *[requirements for plugin, requirements in plugins.items()]
            )
        ),
    },
)



"""
Mixin class containing Table specific methods

To be used by OpenMetadata class
"""
import logging
from typing import List

from metadata.generated.schema.entity.data.location import Location
from metadata.generated.schema.entity.data.table import (
    DataModel,
    Table,
    TableData,
    TableJoins,
    TableProfile,
)
from metadata.ingestion.models.table_queries import TableUsageRequest
from metadata.ingestion.ometa.client import REST

logger = logging.getLogger(__name__)


class OMetaTableMixin:
    """
    OpenMetadata API methods related to Tables.

    To be inherited by OpenMetadata
    """

    client: REST

    def add_location(self, table: Table, location: Location) -> None:
        """
        PUT location for a table

        :param table: Table Entity to update
        :param location: Location Entity to add
        """
        resp = self.client.put(
            f"{self.get_suffix(Table)}/{table.id.__root__}/location",
            data=str(location.id.__root__),
        )

    def ingest_table_sample_data(
        self, table: Table, sample_data: TableData
    ) -> TableData:
        """
        PUT sample data for a table

        :param table: Table Entity to update
        :param sample_data: Data to add
        """
        resp = self.client.put(
            f"{self.get_suffix(Table)}/{table.id.__root__}/sampleData",
            data=sample_data.json(),
        )
        return TableData(**resp["sampleData"])

    def ingest_table_profile_data(
        self, table: Table, table_profile: List[TableProfile]
    ) -> List[TableProfile]:
        """
        PUT profile data for a table

        :param table: Table Entity to update
        :param table_profile: Profile data to add
        """
        for profile in table_profile:
            resp = self.client.put(
                f"{self.get_suffix(Table)}/{table.id.__root__}/tableProfile",
                data=profile.json(),
            )
        return [TableProfile(**t) for t in resp["tableProfile"]]

    def ingest_table_data_model(self, table: Table, data_model: DataModel) -> Table:
        """
        PUT data model for a table

        :param table: Table Entity to update
        :param data_model: Model to add
        """
        resp = self.client.put(
            f"{self.get_suffix(Table)}/{table.id.__root__}/dataModel",
            data=data_model.json(),
        )
        return Table(**resp)

    def publish_table_usage(
        self, table: Table, table_usage_request: TableUsageRequest
    ) -> None:
        """
        POST usage details for a Table

        :param table: Table Entity to update
        :param table_usage_request: Usage data to add
        """
        resp = self.client.post(
            f"/usage/table/{table.id.__root__}", data=table_usage_request.json()
        )
        logger.debug("published table usage {}".format(resp))

    def publish_frequently_joined_with(
        self, table: Table, table_join_request: TableJoins
    ) -> None:
        """
        POST frequently joined with for a table

        :param table: Table Entity to update
        :param table_join_request: Join data to add
        """
        logger.info("table join request {}".format(table_join_request.json()))
        resp = self.client.put(
            f"{self.get_suffix(Table)}/{table.id.__root__}/joins",
            data=table_join_request.json(),
        )
        logger.debug("published frequently joined with {}".format(resp))



# generated by datamodel-codegen:
#   filename:  schema/type/storage.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from enum import Enum
from typing import Any, Union

from pydantic import BaseModel, Field


class Model(BaseModel):
    __root__: Any = Field(..., description='Definitions related to Storage Service')


class StorageServiceType(Enum):
    S3 = 'S3'
    GCS = 'GCS'
    HDFS = 'HDFS'
    ABFS = 'ABFS'


class S3StorageClass(Enum):
    DEEP_ARCHIVE = 'DEEP_ARCHIVE'
    GLACIER = 'GLACIER'
    INTELLIGENT_TIERING = 'INTELLIGENT_TIERING'
    ONEZONE_IA = 'ONEZONE_IA'
    OUTPOSTS = 'OUTPOSTS'
    REDUCED_REDUNDANCY = 'REDUCED_REDUNDANCY'
    STANDARD = 'STANDARD'
    STANDARD_IA = 'STANDARD_IA'


class GcsStorageClass(Enum):
    ARCHIVE = 'ARCHIVE'
    COLDLINE = 'COLDLINE'
    DURABLE_REDUCED_AVAILABILITY = 'DURABLE_REDUCED_AVAILABILITY'
    MULTI_REGIONAL = 'MULTI_REGIONAL'
    NEARLINE = 'NEARLINE'
    REGIONAL = 'REGIONAL'
    STANDARD = 'STANDARD'


class AbfsStorageClass(Enum):
    ARCHIVE = 'ARCHIVE'
    HOT = 'HOT'
    COOL = 'COOL'


class StorageClassType(BaseModel):
    __root__: Union[S3StorageClass, GcsStorageClass, AbfsStorageClass] = Field(
        ..., description='Type of storage class for the storage service'
    )



# generated by datamodel-codegen:
#   filename:  schema/entity/services/databaseService.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from enum import Enum
from typing import Optional

from pydantic import BaseModel, Extra, Field, constr

from ...type import basic, entityHistory, jdbcConnection, schedule


class DatabaseServiceType(Enum):
    BigQuery = 'BigQuery'
    MySQL = 'MySQL'
    Redshift = 'Redshift'
    Snowflake = 'Snowflake'
    Postgres = 'Postgres'
    MSSQL = 'MSSQL'
    Hive = 'Hive'
    Oracle = 'Oracle'
    Athena = 'Athena'
    Presto = 'Presto'
    Trino = 'Trino'
    Vertica = 'Vertica'
    Glue = 'Glue'
    MariaDB = 'MariaDB'
    Druid = 'Druid'


class DatabaseService(BaseModel):
    class Config:
        extra = Extra.forbid

    id: basic.Uuid = Field(
        ..., description='Unique identifier of this database service instance.'
    )
    name: constr(min_length=1, max_length=128) = Field(
        ..., description='Name that identifies this database service.'
    )
    displayName: Optional[str] = Field(
        None, description='Display Name that identifies this database service.'
    )
    serviceType: DatabaseServiceType = Field(
        ...,
        description='Type of database service such as MySQL, BigQuery, Snowflake, Redshift, Postgres...',
    )
    description: Optional[str] = Field(
        None, description='Description of a database service instance.'
    )
    version: Optional[entityHistory.EntityVersion] = Field(
        None, description='Metadata version of the entity.'
    )
    updatedAt: Optional[basic.DateTime] = Field(
        None,
        description='Last update time corresponding to the new version of the entity.',
    )
    updatedBy: Optional[str] = Field(None, description='User who made the update.')
    href: basic.Href = Field(
        ..., description='Link to the resource corresponding to this database service.'
    )
    jdbc: jdbcConnection.JdbcInfo = Field(
        ..., description='JDBC connection information.'
    )
    ingestionSchedule: Optional[schedule.Schedule] = Field(
        None, description='Schedule for running metadata ingestion jobs.'
    )
    changeDescription: Optional[entityHistory.ChangeDescription] = Field(
        None, description='Change that lead to this version of the entity.'
    )



# generated by datamodel-codegen:
#   filename:  schema/api/lineage/addLineage.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Field

from ...type import entityLineage


class AddLineage(BaseModel):
    description: Optional[str] = Field(
        None, description='User provided description of the lineage details.'
    )
    edge: entityLineage.EntitiesEdge = Field(..., description='Lineage edge details.')



# generated by datamodel-codegen:
#   filename:  schema/api/data/createChart.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic import AnyUrl, BaseModel, Field, constr

from ...entity.data import chart
from ...type import entityReference, tagLabel


class CreateChartEntityRequest(BaseModel):
    name: constr(min_length=1, max_length=128) = Field(
        ..., description='Name that identifies this Chart.'
    )
    displayName: Optional[str] = Field(
        None,
        description='Display Name that identifies this Chart. It could be title or label from the source services',
    )
    description: Optional[str] = Field(
        None,
        description='Description of the chart instance. What it has and how to use it.',
    )
    chartType: Optional[chart.ChartType] = None
    chartUrl: Optional[AnyUrl] = Field(
        None, description='Chart URL, pointing to its own Service URL'
    )
    tables: Optional[entityReference.EntityReferenceList] = Field(
        None, description='Link to tables used in this chart.'
    )
    tags: Optional[List[tagLabel.TagLabel]] = Field(
        None, description='Tags for this chart'
    )
    owner: Optional[entityReference.EntityReference] = Field(
        None, description='Owner of this database'
    )
    service: entityReference.EntityReference = Field(
        ..., description='Link to the database service where this database is hosted in'
    )



import json
import logging

from django.shortcuts import render

logger: logging.Logger = logging.getLogger(__name__)

with open("/tmp/datasets.json", "r") as d:
    data = json.load(d)
    datasets = []
    for dataset in data:
        if "table" in dataset:
            datasets.append(dataset["table"])
        elif "dashboard" in dataset:
            datasets.append(dataset["dashboard"])
        elif "topic" in dataset:
            datasets.append(dataset["topic"])
        else:
            logger.info("unrecognized element {}".format(dataset))


def list_page(request):
    template_name = "list.html"
    return render(request, template_name, {"datasets": datasets})


def detail_page(request, fqn):
    template_name = "dataset.html"
    dataset = list(filter(lambda data: data["fullyQualifiedName"] == fqn, datasets))
    if len(dataset) > 0:
        return render(request, template_name, {"dataset": dataset[0]})
    else:
        return render(request, template_name, {"error": "No Data Found"})


def page_not_found_view(request, exception):
    return render(request, "404.html", status=404)



# generated by datamodel-codegen:
#   filename:  schema/entity/teams/team.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Extra, Field, constr

from ...type import basic, entityHistory, entityReference, profile


class TeamName(BaseModel):
    __root__: constr(min_length=1, max_length=128) = Field(
        ...,
        description='A unique name of the team typically the team ID from an identity provider. Example - group Id from LDAP.',
    )


class Team(BaseModel):
    class Config:
        extra = Extra.forbid

    id: basic.Uuid
    name: TeamName
    displayName: Optional[str] = Field(
        None, description="Name used for display purposes. Example 'Data Science team'."
    )
    description: Optional[str] = Field(None, description='Description of the team.')
    version: Optional[entityHistory.EntityVersion] = Field(
        None, description='Metadata version of the entity.'
    )
    updatedAt: Optional[basic.DateTime] = Field(
        None,
        description='Last update time corresponding to the new version of the entity.',
    )
    updatedBy: Optional[str] = Field(None, description='User who made the update.')
    href: basic.Href = Field(
        ..., description='Link to the resource corresponding to this entity.'
    )
    profile: Optional[profile.Profile] = Field(
        None, description='Team profile information.'
    )
    deleted: Optional[bool] = Field(
        None, description='When true the team has been deleted.'
    )
    users: Optional[entityReference.EntityReferenceList] = Field(
        None, description='Users that are part of the team.'
    )
    owns: Optional[entityReference.EntityReferenceList] = Field(
        None, description='List of entities owned by the team.'
    )
    changeDescription: Optional[entityHistory.ChangeDescription] = Field(
        None, description='Change that lead to this version of the entity.'
    )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import textwrap

TABLE_ELASTICSEARCH_INDEX_MAPPING = textwrap.dedent(
    """
     {
    "mappings":{
          "properties": {
            "name": {
              "type":"text"
            },
            "display_name": {
              "type": "text"
            },
            "owner": {
              "type": "text"
            },
            "followers": {
              "type": "keyword"
            },
            "fqdn": {
              "type": "keyword"
            },
            "last_updated_timestamp": {
              "type": "date",
              "format": "epoch_second"
            },
            "description": {
              "type": "text"
            },
            "tier": {
              "type": "keyword"
            },
            "column_names": {
              "type":"text"
            },
            "column_descriptions": {
              "type": "text"
            },
            "tags": {
              "type": "keyword"
            },
            "service": {
              "type": "keyword"
            },
            "service_type": {
              "type": "keyword"
            },
            "service_category": {
              "type": "keyword"
            },
            "entity_type": {
              "type": "keyword"
            },
            "database": {
              "type": "keyword"
            },
            "suggest": {
              "type": "completion"
            },
            "monthly_stats":{
              "type": "long"
            },
            "monthly_percentile_rank":{
              "type": "long"
            },
            "weekly_stats":{
              "type": "long"
            },
            "weekly_percentile_rank":{
              "type": "long"
            },
            "daily_percentile_rank": {
             "type": "long"
            },
            "daily_stats": {
              "type": "long"
            },
            "change_descriptions": {
              "type": "nested",
              "properties": {
                "updatedAt": {
                  "type": "long"
                },
                "updatedBy": {
                  "type": "text"
                },
                "fieldsAdded": {
                  "type": "object",
                  "enabled": false
                },
                "fieldsDeleted": {
                   "type": "object",
                   "enabled": false
                },
                "fieldsUpdated": {
                  "type": "object",
                  "enabled": false
                }
            }
          }
         }
          }
      }
    }
    """
)

TOPIC_ELASTICSEARCH_INDEX_MAPPING = textwrap.dedent(
    """
    {
    "mappings":{
          "properties": {
            "name": {
              "type":"text"
            },
            "display_name": {
              "type": "text"
            },
            "owner": {
              "type": "text"
            },
            "followers": {
              "type": "keyword"
            },
            "fqdn": {
              "type": "keyword"
            },
            "last_updated_timestamp": {
              "type": "date",
              "format": "epoch_second"
            },
            "description": {
              "type": "text"
            },
            "tier": {
              "type": "keyword"
            },
            "tags": {
              "type": "keyword"
            },
            "service": {
              "type": "keyword"
            },
            "service_type": {
              "type": "keyword"
            },
            "service_category": {
              "type": "keyword"
            },
            "entity_type": {
              "type": "keyword"
            },
            "suggest": {
              "type": "completion"
            },
            "change_descriptions": {
              "type": "nested",
              "properties": {
                "updatedAt": {
                  "type": "long"
                },
                "updatedBy": {
                  "type": "text"
                },
                "fieldsAdded": {
                  "type": "object",
                  "enabled": false
                },
                "fieldsDeleted": {
                   "type": "object",
                   "enabled": false
                },
                "fieldsUpdated": {
                  "type": "object",
                  "enabled": false
                }
            }
          }
         }
          }
        }
    }
    """
)

DASHBOARD_ELASTICSEARCH_INDEX_MAPPING = textwrap.dedent(
    """
    {
    "mappings":{
          "properties": {
            "name": {
              "type":"text"
            },
            "display_name": {
              "type": "text"
            },
            "owner": {
              "type": "keyword"
            },
            "fqdn": {
              "type": "keyword"
            },
            "followers": {
              "type": "keyword"
            },
            "last_updated_timestamp": {
              "type": "date",
              "format": "epoch_second"
            },
            "description": {
              "type": "text"
            },
            "chart_names": {
              "type":"text"
            },
            "chart_descriptions": {
              "type": "text"
            },
            "tier": {
              "type": "keyword"
            },
            "tags": {
              "type": "keyword"
            },
            "service": {
              "type": "keyword"
            },
            "service_type": {
              "type": "keyword"
            },
            "service_category": {
              "type": "keyword"
            },
            "entity_type": {
              "type": "keyword"
            },
            "suggest": {
              "type": "completion"
            },
             "monthly_stats":{
              "type": "long"
            },
            "monthly_percentile_rank":{
              "type": "long"
            },
            "weekly_stats":{
              "type": "long"
            },
            "weekly_percentile_rank":{
              "type": "long"
            },
            "daily_percentile_rank": {
             "type": "long"
            },
            "daily_stats": {
              "type": "long"
            },
            "change_descriptions": {
              "type": "nested",
              "properties": {
                "updatedAt": {
                  "type": "long"
                },
                "updatedBy": {
                  "type": "text"
                },
                "fieldsAdded": {
                  "type": "object",
                  "enabled": false
                },
                "fieldsDeleted": {
                   "type": "object",
                   "enabled": false
                },
                "fieldsUpdated": {
                  "type": "object",
                  "enabled": false
                }
            }
          }
         }
          }
        }
    }
    """
)

PIPELINE_ELASTICSEARCH_INDEX_MAPPING = textwrap.dedent(
    """
    {
    "mappings":{
          "properties": {
            "name": {
              "type":"text"
            },
            "display_name": {
              "type": "text"
            },
            "fqdn": {
              "type": "keyword"
            },
            "owner": {
              "type": "keyword"
            },
            "followers": {
              "type": "keyword"
            },
            "last_updated_timestamp": {
              "type": "date",
              "format": "epoch_second"
            },
            "description": {
              "type": "text"
            },
            "task_names": {
              "type":"text"
            },
            "task_descriptions": {
              "type": "text"
            },
            "tier": {
              "type": "keyword"
            },
            "tags": {
              "type": "keyword"
            },
            "service": {
              "type": "keyword"
            },
            "service_type": {
              "type": "keyword"
            },
            "service_category": {
              "type": "keyword"
            },
            "entity_type": {
              "type": "keyword"
            },
            "suggest": {
              "type": "completion"
            },
            "change_descriptions": {
              "type": "nested",
              "properties": {
                "updatedAt": {
                  "type": "long"
                },
                "updatedBy": {
                  "type": "text"
                },
                "fieldsAdded": {
                  "type": "object",
                  "enabled": false
                },
                "fieldsDeleted": {
                   "type": "object",
                   "enabled": false
                },
                "fieldsUpdated": {
                  "type": "object",
                  "enabled": false
                }
            }
          }
         }
      }
    }
    """
)



# generated by datamodel-codegen:
#   filename:  schema/type/entityHistory.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import Any, List, Optional

from pydantic import BaseModel, Extra, Field, confloat


class EntityVersionHistory(BaseModel):
    class Config:
        extra = Extra.forbid

    entityType: str = Field(
        ...,
        description='Entity type, such as `database`, `table`, `dashboard`, for which this version history is produced.',
    )
    versions: List


class EntityVersion(BaseModel):
    __root__: confloat(ge=0.1, multiple_of=0.1) = Field(
        ...,
        description='Metadata version of the entity in the form `Major.Minor`. First version always starts from `0.1` when the entity is created. When the backward compatible changes are made to the entity, only the `Minor` version is incremented - example `1.0` is changed to `1.1`. When backward incompatible changes are made the `Major` version is incremented - example `1.1` to `2.0`.',
    )


class FieldName(BaseModel):
    __root__: str = Field(..., description='Name of the field of an entity.')


class FieldChange(BaseModel):
    class Config:
        extra = Extra.forbid

    name: Optional[FieldName] = Field(
        None, description='Name of the entity field that changed.'
    )
    oldValue: Optional[Any] = Field(
        None,
        description='Previous value of the field. Note that this is a JSON string and use the corresponding field type to deserialize it.',
    )
    newValue: Optional[Any] = Field(
        None,
        description='New value of the field. Note that this is a JSON string and use the corresponding field type to deserialize it.',
    )


class ChangeDescription(BaseModel):
    class Config:
        extra = Extra.forbid

    fieldsAdded: Optional[List[FieldChange]] = Field(
        None, description='Names of fields added during the version changes.'
    )
    fieldsUpdated: Optional[List[FieldChange]] = Field(
        None,
        description='Fields modified during the version changes with old and new values.',
    )
    fieldsDeleted: Optional[List[FieldChange]] = Field(
        None,
        description='Fields deleted during the version changes with old value before deleted.',
    )
    previousVersion: Optional[EntityVersion] = Field(
        None,
        description='When a change did not result in change, this could be same as the current version.',
    )



# generated by datamodel-codegen:
#   filename:  schema/api/teams/createUser.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Extra, Field

from ...entity.teams import user
from ...type import basic, profile


class CreateUserEntityRequest(BaseModel):
    class Config:
        extra = Extra.forbid

    name: user.UserName
    description: Optional[str] = Field(None, description='Used for user biography.')
    displayName: Optional[str] = Field(
        None, description="Name used for display purposes. Example 'FirstName LastName'"
    )
    email: basic.Email
    timezone: Optional[str] = Field(None, description='Timezone of the user')
    isBot: Optional[bool] = Field(
        None,
        description='When true indicates user is a bot with appropriate privileges',
    )
    isAdmin: Optional[bool] = Field(
        False,
        description='When true indicates user is an administrator for the system with superuser privileges',
    )
    profile: Optional[profile.Profile] = None
    teams: Optional[List[basic.Uuid]] = Field(
        None, description='Teams that the user belongs to'
    )
